import json
from logging import fatal

from mm_std import Err, print_json, print_plain
from pydantic import StrictStr

from mm_eth import abi, rpc
from mm_eth.cli import cli_utils


class Config(cli_utils.BaseConfig):
    contract_address: StrictStr
    function_signature: str
    function_args: StrictStr = "[]"
    outputs_types: str | None = None
    node: str


def run(config_path: str, print_config: bool) -> None:
    config = cli_utils.read_config(Config, config_path)
    if print_config:
        print_json(config.model_dump())
        exit(0)

    input_data = abi.encode_function_input_by_signature(
        config.function_signature,
        json.loads(config.function_args.replace("'", '"')),
    )
    res = rpc.eth_call(config.node, config.contract_address, input_data)
    if isinstance(res, Err):
        return fatal(f"error: {res.err}")

    result = res.ok
    if config.outputs_types is not None:
        decode_res = abi.decode_data(_get_types(config.outputs_types), result)
        if len(decode_res) == 1:
            result = decode_res[0]
        else:
            result = str(decode_res)
    print_plain(result)


def _get_types(data: str) -> list[str]:
    return [t.strip() for t in data.split(",") if t.strip()]



from __future__ import annotations

from typing import Any

import rlp
from eth_utils import keccak
from pydantic import BaseModel
from rlp.sedes import Binary, big_endian_int, binary
from web3 import Web3
from web3.auto import w3

from mm_eth.utils import hex_to_bytes


class SignedTx(BaseModel):
    tx_hash: str
    raw_tx: str


class RPLTransaction(rlp.Serializable):  # type: ignore[misc]
    fields = [  # noqa: RUF012
        ("nonce", big_endian_int),
        ("gas_price", big_endian_int),
        ("gas", big_endian_int),
        ("to", Binary.fixed_length(20, allow_empty=True)),
        ("value", big_endian_int),
        ("data", binary),
        ("v", big_endian_int),
        ("r", big_endian_int),
        ("s", big_endian_int),
    ]

    @staticmethod
    def new_tx(
        *,
        nonce: int,
        gas_price: int,
        gas: int,
        v: int,
        r: str,
        s: str,
        data: str | None = None,
        value: int | None = None,
        to: str | None = None,
    ) -> RPLTransaction:
        if to:
            to = hex_to_bytes(to)  # type:ignore
        if data:
            data = hex_to_bytes(data)  # type:ignore
        if not value:
            value = 0
        r = int(r, 16)  # type:ignore
        s = int(s, 16)  # type:ignore
        return RPLTransaction(nonce, gas_price, gas, to, value, data, v, r, s)


class DecodedRawTx(BaseModel):
    tx_hash: str
    from_: str
    to: str | None
    nonce: int
    gas: int
    gas_price: int
    value: int
    data: str
    chain_id: int
    r: str
    s: str
    v: int


def encode_raw_tx_with_signature(
    *,
    nonce: int,
    gas_price: int,
    gas: int,
    v: int,
    r: str,
    s: str,
    data: str | None = None,
    value: int | None = None,
    to: str | None = None,
) -> str:
    tx = RPLTransaction.new_tx(nonce=nonce, gas_price=gas_price, gas=gas, v=v, r=r, s=s, data=data, value=value, to=to)
    return Web3.to_hex(rlp.encode(tx))


def sign_legacy_tx(
    *,
    nonce: int,
    gas_price: int,
    gas: int,
    private_key: str,
    chain_id: int,
    data: str | None = None,
    value: int | None = None,
    to: str | None = None,
) -> SignedTx:
    tx: dict[str, Any] = {"gas": gas, "gasPrice": gas_price, "nonce": nonce, "chainId": chain_id}
    if to:
        tx["to"] = Web3.to_checksum_address(to)
    if value:
        tx["value"] = value
    if data:
        tx["data"] = data

    signed = w3.eth.account.sign_transaction(tx, private_key)
    return SignedTx(tx_hash=signed.hash.hex(), raw_tx=signed.rawTransaction.hex())


def sign_tx(
    *,
    nonce: int,
    max_fee_per_gas: int,
    max_priority_fee_per_gas: int,
    gas: int,
    private_key: str,
    chain_id: int,
    data: str | None = None,
    value: int | None = None,
    to: str | None = None,
) -> SignedTx:
    tx: dict[str, Any] = {
        "type": "0x2",
        "gas": gas,
        "maxFeePerGas": max_fee_per_gas,
        "maxPriorityFeePerGas": max_priority_fee_per_gas,
        "nonce": nonce,
        "chainId": chain_id,
    }
    if value:
        tx["value"] = value
    if data:
        tx["data"] = data
    if to:
        tx["to"] = Web3.to_checksum_address(to)

    signed = w3.eth.account.sign_transaction(tx, private_key)
    return SignedTx(tx_hash=signed.hash.hex(), raw_tx=signed.rawTransaction.hex())


def decode_raw_tx(raw_tx: str) -> DecodedRawTx:
    tx: Any = rlp.decode(hex_to_bytes(raw_tx), RPLTransaction)
    tx_hash = Web3.to_hex(keccak(hex_to_bytes(raw_tx)))
    from_ = w3.eth.account.recover_transaction(raw_tx)
    to = Web3.to_checksum_address(tx.to) if tx.to else None
    data = Web3.to_hex(tx.data)
    r = hex(tx.r)
    s = hex(tx.s)
    chain_id = (tx.v - 35) // 2 if tx.v % 2 else (tx.v - 36) // 2
    return DecodedRawTx(
        tx_hash=tx_hash,
        from_=from_,
        to=to,
        data=data,
        chain_id=chain_id,
        r=r,
        s=s,
        v=tx.v,
        gas=tx.gas,
        gas_price=tx.gas_price,
        value=tx.value,
        nonce=tx.nonce,
    )



from __future__ import annotations

import string
from collections.abc import Sequence
from dataclasses import dataclass

import eth_abi
import eth_utils
from eth_typing import HexStr
from eth_utils import to_checksum_address, to_hex
from mm_std import Err, Ok, Result

from mm_eth import rpc
from mm_eth.rpc import Log
from mm_eth.tx import SignedTx, sign_legacy_tx, sign_tx
from mm_eth.types import Nodes, Proxies
from mm_eth.utils import hex_str_to_int, hex_to_bytes, log_topic_to_address

TRANSFER_METHOD = "0xa9059cbb"
TRANSFER_TOPIC = HexStr("0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef")


USDT_MAINNET_ADDRESS = "0xdac17f958d2ee523a2206206994597c13d831ec7"


@dataclass
class TransferEventLog:
    token_address: str
    from_address: str
    to_address: str
    value: int
    tx_hash: str
    block_number: int
    log_index: int

    @staticmethod
    def from_log(log: Log) -> Result[TransferEventLog]:
        try:
            return Ok(
                TransferEventLog(
                    token_address=log.address,
                    from_address=log_topic_to_address(log.topics[1]),
                    to_address=log_topic_to_address(log.topics[2]),
                    value=int(log.data, 16),
                    tx_hash=log.transaction_hash,
                    block_number=log.block_number,
                    log_index=log.log_index,
                ),
            )
        except Exception as err:
            return Err(err)


def get_balance(
    rpc_urls: Nodes,
    token_address: str,
    user_address: str,
    timeout: int = 10,
    proxies: Proxies = None,
    attempts: int = 1,
) -> Result[int]:
    data = "0x70a08231000000000000000000000000" + user_address[2:]
    return rpc.rpc_call(
        nodes=rpc_urls,
        method="eth_call",
        params=[{"to": token_address, "data": data}, "latest"],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    ).and_then(hex_str_to_int)


def get_name(rpc_urls: Nodes, address: str, timeout: int = 10, proxies: Proxies = None, attempts: int = 1) -> Result[str]:
    return rpc.rpc_call(
        nodes=rpc_urls,
        method="eth_call",
        params=[{"to": address, "data": "0x06fdde03"}, "latest"],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    ).and_then(_normalize_str)


def get_symbol(rpc_urls: Nodes, address: str, timeout: int = 10, proxies: Proxies = None, attempts: int = 1) -> Result[str]:
    return rpc.rpc_call(
        nodes=rpc_urls,
        method="eth_call",
        params=[{"to": address, "data": "0x95d89b41"}, "latest"],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    ).and_then(_normalize_str)


def get_decimals(rpc_urls: Nodes, address: str, timeout: int = 10, proxies: Proxies = None, attempts: int = 1) -> Result[int]:
    res = rpc.rpc_call(
        nodes=rpc_urls,
        method="eth_call",
        params=[{"to": address, "data": "0x313ce567"}, "latest"],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    )
    if isinstance(res, Err):
        return res

    try:
        if res.ok == "0x":
            return Err("no_decimals", data=res.data)

        if len(res.ok) > 66:
            result = eth_utils.to_int(hexstr=res.ok[0:66])
        else:
            result = eth_utils.to_int(hexstr=res.ok)
        return Ok(result, data=res.data)

    except Exception as e:
        return Err(f"exception: {e}", data=res.data)


def encode_transfer_input_data(recipient: str, value: int) -> str:
    recipient = to_checksum_address(recipient)
    input_data = hex_to_bytes(TRANSFER_METHOD) + eth_abi.encode(["address", "uint256"], [recipient, value])
    return to_hex(input_data)


def transfer_token_legacy(
    *,
    rpc_urls: str | Sequence[str],
    token_address: str,
    recipient_address: str,
    value: int,
    nonce: int,
    gas_price: int,
    gas_limit: int,
    private_key: str,
    chain_id: int,
    timeout: int = 10,
    proxies: Proxies = None,
    attempts: int = 1,
) -> Result[str]:
    input_data = encode_transfer_input_data(recipient_address, value)
    signed_tx = sign_legacy_tx(
        nonce=nonce,
        gas_price=gas_price,
        gas=gas_limit,
        private_key=private_key,
        chain_id=chain_id,
        data=input_data,
        to=token_address,
    )
    return rpc.eth_send_raw_transaction(rpc_urls, signed_tx.raw_tx, timeout=timeout, proxies=proxies, attempts=attempts)


def transfer_token(
    *,
    rpc_urls: Nodes,
    token_address: str,
    recipient_address: str,
    value: int,
    nonce: int,
    max_fee_per_gas: int,
    max_priority_fee_per_gas: int,
    gas_limit: int,
    private_key: str,
    chain_id: int,
    timeout: int = 10,
    proxies: Proxies = None,
    attempts: int = 1,
) -> Result[str]:
    input_data = encode_transfer_input_data(recipient_address, value)
    signed_tx = sign_tx(
        nonce=nonce,
        max_fee_per_gas=max_fee_per_gas,
        max_priority_fee_per_gas=max_priority_fee_per_gas,
        gas=gas_limit,
        private_key=private_key,
        chain_id=chain_id,
        data=input_data,
        to=token_address,
    )
    return rpc.eth_send_raw_transaction(rpc_urls, signed_tx.raw_tx, timeout=timeout, proxies=proxies, attempts=attempts)


def sign_transfer_tx(
    *,
    token_address: str,
    recipient_address: str,
    value: int,
    nonce: int,
    max_fee_per_gas: int,
    max_priority_fee_per_gas: int,
    gas_limit: int,
    private_key: str,
    chain_id: int,
) -> SignedTx:
    input_data = encode_transfer_input_data(recipient_address, value)
    return sign_tx(
        nonce=nonce,
        max_fee_per_gas=max_fee_per_gas,
        max_priority_fee_per_gas=max_priority_fee_per_gas,
        gas=gas_limit,
        private_key=private_key,
        chain_id=chain_id,
        data=input_data,
        to=token_address,
    )


def get_transfer_event_logs(
    rpc_urls: Nodes,
    token_address: str | None,
    from_block: int,
    to_block: int,
    timeout: int = 10,
    proxies: Proxies = None,
    attempts: int = 1,
) -> Result[list[TransferEventLog]]:
    res = rpc.eth_get_logs(
        rpc_urls,
        address=token_address,
        topics=[TRANSFER_TOPIC],
        from_block=from_block,
        to_block=to_block,
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    )
    if isinstance(res, Err):
        return res
    result: list[TransferEventLog] = []
    for log in res.ok:
        event_log_res = TransferEventLog.from_log(log)
        if isinstance(event_log_res, Ok):
            result.append(event_log_res.ok)
    return Ok(result, data=res.data)


def _normalize_str(value: str) -> Ok[str]:
    return Ok("".join(filter(lambda x: x in string.printable, eth_utils.to_text(hexstr=value))).strip())



from mm_std import Err, Ok, print_json, print_plain

from mm_eth import erc20, rpc
from mm_eth.cli import cli_helpers
from mm_eth.cli.cli import PrintFormat
from mm_eth.cli.cli_utils import public_rpc_url
from mm_eth.utils import from_wei_str


def run(rpc_url: str, wallet_address: str, token_address: str | None, wei: bool, print_format: PrintFormat) -> None:
    rpc_url = public_rpc_url(rpc_url)
    json_result: dict[str, object] = {}

    # nonce
    nonce = rpc.eth_get_transaction_count(rpc_url, wallet_address).ok_or_err()
    print_plain(f"nonce: {nonce}", print_format)
    json_result["nonce"] = nonce

    # balance
    balance_res = rpc.eth_get_balance(rpc_url, wallet_address)
    if isinstance(balance_res, Ok):
        if wei:
            balance = str(balance_res.ok)
        else:
            balance = from_wei_str(balance_res.ok, "eth")
    else:
        balance = balance_res.err
    print_plain(f"eth_balance: {balance}", print_format)
    json_result["eth_balance"] = balance

    if token_address is not None:
        # token decimal
        decimals_res = erc20.get_decimals(rpc_url, token_address)
        if isinstance(decimals_res, Err):
            cli_helpers.fatal(f"can't get token decimals: {decimals_res.err}")
        decimals = decimals_res.ok
        print_plain(f"token_decimal: {decimals}", print_format)
        json_result["token_decimal"] = decimals

        # token symbol
        symbol_res = erc20.get_symbol(rpc_url, token_address)
        if isinstance(symbol_res, Err):
            cli_helpers.fatal(f"can't get token symbol: {symbol_res.err}")
        symbol = symbol_res.ok
        print_plain(f"token_symbol: {symbol}", print_format)
        json_result["token_symbol"] = symbol

        # token balance
        balance_res = erc20.get_balance(rpc_url, token_address, wallet_address)
        if isinstance(balance_res, Err):
            cli_helpers.fatal(f"can't get token balance: {balance_res.err}")
        if wei:
            balance = str(balance_res.ok)
        else:
            balance = from_wei_str(balance_res.ok, "t", decimals=decimals)
        print_plain(f"token_balance: {balance}", print_format)
        json_result["token_balance"] = balance

    print_json(json_result, print_format)



import json
from typing import cast

from mm_std import fatal, hr, print_console, str_starts_with_any
from rich import print_json

from mm_eth.cli.cli_utils import public_rpc_url


def run(rpc_url: str, method: str, params: str, hex2dec: bool) -> None:
    rpc_url = public_rpc_url(rpc_url)
    if not method:
        return list_all_methods()
    if not str_starts_with_any(rpc_url, ["http://", "https://"]):
        fatal(f"invalid rpc_url: {rpc_url}")
    params = params.replace("'", '"')
    data = {"jsonrpc": "2.0", "method": method, "params": parse_method_params(method, params), "id": 1}
    res = hr(rpc_url, method="POST", params=data, json_params=True)
    if res.json:
        print_json(data=res.json)
        result_value: str = res.json.get("result", "")
        if hex2dec and result_value.startswith(("0x", "0X")):
            print_console("hex2dec", int(result_value, 16))
    else:
        fatal(str(res))


def parse_method_params(method: str, params_str: str) -> list[object]:
    if params_str.startswith("["):
        params = json.loads(params_str)
    else:
        params = params_str.split()
    if method in ["eth_getBalance", "eth_getTransactionCount", "eth_getCode"] and len(params) == 1:
        params.append("latest")
    return cast(list[object], params)


def list_all_methods() -> None:
    all_methods = """
web3_clientVersion
web3_sha3
net_version
net_listening
net_peerCount
eth_protocolVersion
eth_syncing
eth_chainId
eth_gasPrice
eth_accounts
eth_blockNumber
eth_getBalance
eth_getStorageAt
eth_getTransactionCount
eth_getBlockTransactionCountByHash
eth_getBlockTransactionCountByNumber
eth_getUncleCountByBlockHash
eth_getUncleCountByBlockNumber
eth_getCode
eth_sign
eth_signTransaction
eth_sendTransaction
eth_sendRawTransaction
eth_call
eth_estimateGas
eth_getBlockByHash
eth_getBlockByNumber
eth_getTransactionByHash
eth_getTransactionByBlockHashAndIndex
eth_getTransactionByBlockNumberAndIndex
eth_getTransactionReceipt
eth_getUncleByBlockHashAndIndex
eth_getUncleByBlockNumberAndIndex
eth_newFilter
eth_newBlockFilter
eth_newPendingTransactionFilter
eth_uninstallFilter
eth_getFilterChanges
eth_getFilterLogs
eth_getLogs
    """.strip()
    print_console(all_methods)



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Literal, cast

import websocket
from mm_std import Err, Ok, Result, hr, random_choice
from pydantic import BaseModel
from web3.types import BlockIdentifier

from mm_eth.types import Nodes, Proxies
from mm_eth.utils import hex_str_to_int, random_node, random_proxy


@dataclass
class TxReceipt:
    tx_hash: str
    tx_index: int
    block_number: int
    from_address: str
    to_address: str | None
    contract_address: str | None
    status: int | None


@dataclass
class Log:
    address: str
    block_hash: str
    block_number: int
    data: str
    log_index: int
    removed: bool
    topics: list[str]
    transaction_hash: str
    transaction_index: int

    @classmethod
    def from_json_rpc_dict(cls, data: dict[str, Any]) -> Result[Log]:
        try:
            return Ok(
                Log(
                    address=data["address"],
                    block_hash=data["blockHash"],
                    block_number=int(data["blockNumber"], 16),
                    data=data["data"],
                    log_index=int(data["logIndex"], 16),
                    removed=data["removed"],
                    topics=data["topics"],
                    transaction_hash=data["transactionHash"],
                    transaction_index=int(data["transactionIndex"], 16),
                ),
            )
        except Exception as err:
            return Err(f"exception: {err}")


class TxData(BaseModel):
    block_number: int | None  # for pending tx it can be none
    from_: str
    to: str | None
    gas: int
    gas_price: int
    value: int
    hash: str
    input: str
    nonce: int
    v: int
    r: str
    s: str


def rpc_call(
    *,
    nodes: Nodes,
    method: str,
    params: list[object],
    id_: int = 1,
    timeout: int = 10,
    proxies: Proxies = None,
    attempts: int = 1,
) -> Result[Any]:
    data = {"jsonrpc": "2.0", "method": method, "params": params, "id": id_}
    res: Result[Any] = Err("not started yet")
    for _ in range(attempts):
        node = random_node(nodes)
        if node.startswith("http"):
            res = _http_call(node, data, timeout, random_proxy(proxies))
        else:
            res = _ws_call(node, data, timeout)
        if isinstance(res, Ok):
            return res
    return res


def _http_call(node: str, data: dict[str, object], timeout: int, proxy: str | None) -> Result[Any]:
    res = hr(node, method="POST", proxy=proxy, timeout=timeout, params=data, json_params=True)
    try:
        if res.is_error():
            return res.to_err_result()

        err = res.json.get("error", {}).get("message", "")
        if err:
            return res.to_err_result(f"service_error: {err}")
        if "result" in res.json:
            return res.to_ok_result(res.json["result"])

        return res.to_err_result("unknown_response")
    except Exception as err:
        return res.to_err_result(f"exception: {err}")


def _ws_call(node: str, data: dict[str, object], timeout: int) -> Result[Any]:
    try:
        ws = websocket.create_connection(node, timeout=timeout)
        ws.send(json.dumps(data))
        response = json.loads(ws.recv())
        ws.close()
        err = response.get("error", {}).get("message", "")
        if err:
            return Err(f"service_error: {err}")
        if "result" in response:
            return Ok(response["result"])
        return Err(f"unknown_response: {response}")
    except TimeoutError:
        return Err("timeout")
    except Exception as err:
        return Err(f"exception: {err}")


def eth_block_number(rpc_urls: Nodes, timeout: int = 10, proxies: Proxies = None, attempts: int = 1) -> Result[int]:
    return rpc_call(
        nodes=rpc_urls,
        method="eth_blockNumber",
        params=[],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    ).and_then(hex_str_to_int)


def eth_chain_id(rpc_urls: Nodes, timeout: int = 10, proxies: Proxies = None, attempts: int = 1) -> Result[int]:
    return rpc_call(
        nodes=rpc_urls,
        method="eth_chainId",
        params=[],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    ).and_then(hex_str_to_int)


def net_peer_count(rpc_urls: Nodes, timeout: int = 10, proxies: Proxies = None, attempts: int = 1) -> Result[int]:
    return rpc_call(
        nodes=rpc_urls,
        method="net_peerCount",
        params=[],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    ).and_then(hex_str_to_int)


def web3_client_version(rpc_urls: Nodes, timeout: int = 10, proxies: Proxies = None, attempts: int = 1) -> Result[str]:
    return rpc_call(
        nodes=rpc_urls,
        method="web3_clientVersion",
        params=[],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    )


def net_version(nodes: Nodes, timeout: int = 10, proxies: Proxies = None, attempts: int = 1) -> Result[str]:
    return rpc_call(nodes=nodes, method="net_version", params=[], timeout=timeout, proxies=proxies, attempts=attempts)


def eth_get_code(rpc_urls: Nodes, address: str, timeout: int = 10, proxies: Proxies = None, attempts: int = 1) -> Result[str]:
    return rpc_call(
        nodes=rpc_urls,
        method="eth_getCode",
        params=[address, "latest"],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    )


def eth_send_raw_transaction(
    rpc_urls: Nodes,
    raw_tx: str,
    timeout: int = 10,
    proxies: Proxies = None,
    attempts: int = 1,
) -> Result[str]:
    return rpc_call(
        nodes=rpc_urls,
        method="eth_sendRawTransaction",
        params=[raw_tx],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    )


def eth_get_balance(rpc_urls: Nodes, address: str, timeout: int = 10, proxies: Proxies = None, attempts: int = 1) -> Result[int]:
    return rpc_call(
        nodes=rpc_urls,
        method="eth_getBalance",
        params=[address, "latest"],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    ).and_then(hex_str_to_int)


def eth_get_transaction_count(
    rpc_urls: Nodes,
    address: str,
    timeout: int = 10,
    proxies: Proxies = None,
    attempts: int = 1,
) -> Result[int]:
    return rpc_call(
        nodes=rpc_urls,
        method="eth_getTransactionCount",
        params=[address, "latest"],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    ).and_then(hex_str_to_int)


def eth_get_block_by_number(
    rpc_urls: Nodes,
    block_number: BlockIdentifier,
    full_transaction: bool = False,
    timeout: int = 10,
    proxies: Proxies = None,
    attempts: int = 1,
) -> Result[dict[str, Any]]:
    return rpc_call(
        nodes=rpc_urls,
        method="eth_getBlockByNumber",
        params=[hex(block_number) if isinstance(block_number, int) else block_number, full_transaction],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    )


def eth_get_logs(
    rpc_urls: Nodes,
    *,
    address: str | None = None,
    topics: list[str] | None = None,
    from_block: BlockIdentifier | None = None,
    to_block: BlockIdentifier | None = None,
    timeout: int = 10,
    proxies: Proxies = None,
    attempts: int = 1,
) -> Result[list[Log]]:
    params: dict[str, object] = {}
    if address:
        params["address"] = address
    if isinstance(from_block, int):
        params["fromBlock"] = hex(from_block)
    else:
        params["fromBlock"] = "earliest"
    if isinstance(to_block, int):
        params["toBlock"] = hex(to_block)
    if topics:
        params["topics"] = topics

    res = rpc_call(nodes=rpc_urls, method="eth_getLogs", params=[params], proxies=proxies, attempts=attempts, timeout=timeout)
    if isinstance(res, Err):
        return res

    result: list[Log] = []
    for log_data in res.ok:
        log_res = Log.from_json_rpc_dict(log_data)
        if isinstance(log_res, Err):
            return Err(log_res.err, data=res.data)
        result.append(log_res.ok)
    return Ok(result, data=res.data)


def eth_get_transaction_receipt(
    rpc_urls: Nodes,
    tx_hash: str,
    timeout: int = 10,
    proxies: Proxies = None,
    attempts: int = 1,
) -> Result[TxReceipt]:
    res = rpc_call(
        nodes=rpc_urls,
        method="eth_getTransactionReceipt",
        params=[tx_hash],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    )
    if isinstance(res, Err):
        return res

    if res.ok is None:
        return Err("no_receipt", data=res.data)

    try:
        status = None
        receipt = cast(dict[str, Any], res.ok)
        if "status" in receipt:
            status = int(receipt["status"], 16)
        return Ok(
            TxReceipt(
                tx_hash=tx_hash,
                tx_index=int(receipt["transactionIndex"], 16),
                block_number=int(receipt["blockNumber"], 16),
                from_address=receipt["from"],
                to_address=receipt.get("to"),
                contract_address=receipt.get("contractAddress"),
                status=status,
            ),
            data=res.data,
        )
    except Exception as err:
        return Err(f"exception: {err}", data=res.data)


def eth_get_transaction_by_hash(
    rpc_urls: Nodes,
    tx_hash: str,
    timeout: int = 10,
    proxies: Proxies = None,
    attempts: int = 1,
) -> Result[TxData]:
    res = rpc_call(
        nodes=rpc_urls,
        method="eth_getTransactionByHash",
        params=[tx_hash],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    )
    if isinstance(res, Err):
        return res
    if res.ok is None:
        return Err("not_found", data=res.data)

    try:
        tx = res.ok
        return Ok(
            TxData(
                block_number=int(tx["blockNumber"], 16) if tx["blockNumber"] is not None else None,
                from_=tx["from"],
                to=tx.get("to"),
                gas=int(tx["gas"], 16),
                gas_price=int(tx["gasPrice"], 16),
                value=int(tx["value"], 16),
                nonce=int(tx["nonce"], 16),
                input=tx["input"],
                hash=tx_hash,
                v=int(tx["v"], 16),
                r=tx.get("r"),
                s=tx.get("s"),
            ),
            data=res.data,
        )

    except Exception as err:
        return Err(f"exception: {err}", data=res.data)


def eth_call(
    rpc_urls: Nodes,
    to: str,
    data: str,
    timeout: int = 10,
    proxies: Proxies = None,
    attempts: int = 1,
) -> Result[str]:
    return rpc_call(
        nodes=rpc_urls,
        method="eth_call",
        params=[{"to": to, "data": data}, "latest"],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    )


def eth_estimate_gas(
    rpc_urls: Nodes,
    from_: str,
    to: str | None = None,
    value: int | None = 0,
    data: str | None = None,
    type_: Literal["0x0", "0x2"] | None = None,
    timeout: int = 10,
    proxies: Proxies = None,
    attempts: int = 1,
) -> Result[int]:
    params: dict[str, Any] = {"from": from_}
    if to:
        params["to"] = to
    if data:
        params["data"] = data
    if value:
        params["value"] = hex(value)
    if type_:
        params["type"] = type_
    return rpc_call(
        nodes=rpc_urls,
        method="eth_estimateGas",
        params=[params],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    ).and_then(hex_str_to_int)


def eth_gas_price(rpc_urls: Nodes, timeout: int = 10, proxies: Proxies = None, attempts: int = 1) -> Result[int]:
    return rpc_call(
        nodes=rpc_urls,
        method="eth_gasPrice",
        params=[],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    ).and_then(hex_str_to_int)


def eth_syncing(rpc_urls: Nodes, timeout: int = 10, proxies: Proxies = None, attempts: int = 1) -> Result[bool | dict[str, int]]:
    res = rpc_call(nodes=rpc_urls, method="eth_syncing", params=[], timeout=timeout, proxies=proxies, attempts=attempts)
    if isinstance(res, Err):
        return res

    if isinstance(res.ok, dict):
        result = {}
        for k, v in res.ok.items():
            if v:
                result[k] = int(v, 16)
            else:
                result[k] = v
        if result.get("currentBlock", None) and result.get("highestBlock", None):
            result["remaining"] = result["highestBlock"] - result["currentBlock"]
        return Ok(result, res.data)

    return res


def get_tx_status(rpc_urls: Nodes, tx_hash: str, timeout: int = 5, proxies: Proxies = None, attempts: int = 5) -> Result[int]:
    res: Result[int] = Err("not started yet")
    for _ in range(attempts):
        node = cast(str, random_choice(rpc_urls))
        cast(str | None, random_choice(proxies))
        receipt_res = eth_get_transaction_receipt(node, tx_hash, timeout, proxies=proxies, attempts=1)
        if isinstance(receipt_res, Err) and receipt_res.err == "no_receipt":
            return receipt_res
        if isinstance(receipt_res, Ok) and receipt_res.ok.status is None:
            return Err("no_status", data=res.data)

        if isinstance(receipt_res, Ok):
            return Ok(cast(int, receipt_res.ok.status), data=receipt_res.data)
        res = receipt_res

    return res


def get_base_fee_per_gas(rpc_urls: Nodes, timeout: int = 5, proxies: Proxies = None, attempts: int = 5) -> Result[int]:
    res = eth_get_block_by_number(rpc_urls, "latest", False, timeout=timeout, proxies=proxies, attempts=attempts)
    if isinstance(res, Err):
        return res
    if "baseFeePerGas" in res.ok:
        return Ok(int(res.ok["baseFeePerGas"], 16), data=res.data)
    return Err("no_base_fee_per_gas", data=res.data)



import yaml
from mm_std import print_json
from pydantic import StrictStr

from mm_eth import account, deploy
from mm_eth.cli import cli_helpers, cli_utils
from mm_eth.cli.cli_utils import BaseConfig


class Config(BaseConfig):
    private_key: StrictStr
    nonce: int | None = None
    gas: StrictStr
    max_fee_per_gas: str
    max_priority_fee_per_gas: str
    value: str | None = None
    contract_bin: StrictStr
    constructor_types: StrictStr = "[]"
    constructor_values: StrictStr = "[]"
    chain_id: int
    node: str


def run(config_path: str, *, print_config: bool) -> None:
    config = cli_utils.read_config(Config, config_path)
    if print_config:
        print_json(config.model_dump(exclude={"private_key"}))
        exit(0)

    constructor_types = yaml.full_load(config.constructor_types)
    constructor_values = yaml.full_load(config.constructor_values)

    sender_address = account.private_to_address(config.private_key)
    if sender_address is None:
        cli_utils.fatal("private address is wrong")

    nonce = cli_helpers.get_nonce(config.node, sender_address)
    if nonce is None:
        cli_utils.fatal("can't get nonce")

    deploy.get_deploy_contract_data(config.contract_bin, constructor_types, constructor_values)



from pathlib import Path

from mm_std import print_plain


def run(command: str) -> None:
    command = command.replace("-", "_")
    example_file = Path(Path(__file__).parent.absolute(), "../config_examples", f"{command}.yml")
    print_plain(example_file.read_text())



import json

from mm_std import print_plain

from mm_eth import abi


def run(function_signature: str, args_str: str) -> None:
    args_str = args_str.replace("'", '"')
    print_plain(abi.encode_function_input_by_signature(function_signature, json.loads(args_str)))



from mm_std import Err, Ok, Result, hr

from mm_eth.account import private_to_address


def read_keys_from_vault(keys_url: str, token: str) -> Result[dict[str, str]]:
    data = None
    try:
        # keys_url example, https://vault.site.com:8200/v1/kv/keys1
        res = hr(keys_url, headers={"X-Vault-Token": token})
        data = res.json
        return Ok(res.json["data"], data=data)
    except Exception as e:
        return Err(f"exception: {e}", data=data)


def set_keys_from_vault(keys_url: str, token: str, private_keys: list[str], verify_tls: bool = True) -> Result[bool]:
    """It works with KV version=1 only!!!"""  # TODO: check that keys_url is kv1 version and error if it's kv2
    data = None
    try:
        # keys_url example, https://vault.site.com:8200/v1/kv/keys1
        keys: dict[str, str] = {}
        for private_key in private_keys:
            address = private_to_address(private_key)
            if address is None:
                raise ValueError("there is a wrong private key")
            keys[address] = private_key

        res = hr(keys_url, method="post", headers={"X-Vault-Token": token}, params=keys, verify=verify_tls)
        data = res.json
        if res.code == 204:
            return Ok(res.code == 204, data=data)
        elif res.code == 403:
            return Err("permission denied", data=data)
        else:
            return Err(res.error or "error", data=data)
    except Exception as e:
        return Err(f"exception: {e}", data=data)



from mm_eth import account


def test_generate_mnemonic():
    assert len(account.generate_mnemonic().split()) == 24
    assert len(account.generate_mnemonic(12).split()) == 12
    assert account.generate_mnemonic() != account.generate_mnemonic()


def test_generate_accounts():
    assert len(account.generate_accounts(account.generate_mnemonic(), limit=17)) == 17


def test_to_checksum_address():
    address = "0x46246a9e6B931EE2C60a525455c01689bA8eb2Ae"
    assert account.to_checksum_address("0x46246a9e6b931ee2c60a525455c01689ba8eb2ae") == address


def test_private_to_address():
    address = "0x46246a9e6B931EE2C60a525455c01689bA8eb2Ae"
    assert account.private_to_address("0xbc2a0bb29ed04fd94cb413a4483e56187e6faf13c2f6f4ab4ec0fa5bef8fd128") == address.lower()
    assert account.private_to_address("123") is None


def test_is_private_key():
    assert account.is_private_key("0xd17e3e15fd28dea2825073d08ab8b7320555759e5639d889d7b4b314c49743a0")
    assert account.is_private_key("d17e3e15fd28dea2825073d08ab8b7320555759e5639d889d7b4b314c49743a0")
    assert not account.is_private_key("17e3e15fd28dea2825073d08ab8b7320555759e5639d889d7b4b314c49743a0")
    assert not account.is_private_key("d17e3e15fd28dea2825073d08ab8b7320555759e5639d889d7b4b314c49743a09999999")
    assert not account.is_private_key("qwe")
    assert not account.is_private_key("")



from decimal import Decimal

import pytest
from mm_eth import utils


def test_to_wei():
    assert utils.to_wei(123) == 123
    assert utils.to_wei(Decimal("123")) == 123
    assert utils.to_wei("11navax") == 11000000000
    assert utils.to_wei("12.1t", decimals=6) == 12.1 * 10**6

    with pytest.raises(ValueError):
        utils.to_wei(Decimal("123.1"))
    with pytest.raises(ValueError):
        utils.to_wei("10t")


def test_to_token_wei():
    assert utils.to_token_wei("12.3t", 6) == 12300000
    assert utils.to_token_wei("12.3 t", 6) == 12300000


def test_from_wei():
    assert utils.from_wei(123000000000000000, "ether") == Decimal("0.123")
    assert utils.from_wei(0, "ether") == Decimal("0")
    assert utils.from_wei(12.1 * 10**6, "t", decimals=6) == Decimal("12.1")


def test_from_wei_str():
    assert utils.from_wei_str(123000000000000000, "ether", print_unit_name=False) == "0.123"
    assert utils.from_wei_str(123000000000000000, "ether") == "0.123eth"
    assert utils.from_wei_str(123000000000000000, "ether", round_ndigits=2) == "0.12eth"
    assert utils.from_wei_str(123000000000000000, "t", round_ndigits=2, decimals=18) == "0.12t"
    assert utils.from_wei_str(0, "ether", print_unit_name=False) == "0"
    assert utils.from_wei_str(12.1 * 10**6, "t", decimals=6, print_unit_name=False) == "12.1"


def test_to_checksum_address(address_0):
    assert utils.to_checksum_address(address_0.lower()) == address_0


def test_truncate_hex_str(address_0, private_0):
    assert utils.truncate_hex_str(address_0) == "0x10fd...6623"
    assert utils.truncate_hex_str(private_0, digits=3, replace_str="****") == "0x7bb****632"



from datetime import datetime

from mm_std import CHROME_USER_AGENT, Result, hr
from pydantic import BaseModel, Field


class SearchResult(BaseModel):
    class Node(BaseModel):
        id: str
        host: str
        port: int
        client: str
        client_version: str = Field(..., alias="clientVersion")
        os: str
        last_update: datetime = Field(..., alias="lastUpdate")
        country: str
        in_sync: int | None = Field(None, alias="inSync")
        isp: str

    draw: int
    records_total: int = Field(..., alias="recordsTotal")
    records_filtered: int = Field(..., alias="recordsFiltered")
    data: list[Node]


def search_nodes(offset: int = 0, proxy: str | None = None) -> Result[SearchResult]:
    url = f"https://www.ethernodes.org/data?draw=1&columns%5B0%5D%5Bdata%5D=id&columns%5B0%5D%5Bname%5D=&columns%5B0%5D%5Bsearchable%5D=true&columns%5B0%5D%5Borderable%5D=true&columns%5B0%5D%5Bsearch%5D%5Bvalue%5D=&columns%5B0%5D%5Bsearch%5D%5Bregex%5D=false&columns%5B1%5D%5Bdata%5D=host&columns%5B1%5D%5Bname%5D=&columns%5B1%5D%5Bsearchable%5D=true&columns%5B1%5D%5Borderable%5D=true&columns%5B1%5D%5Bsearch%5D%5Bvalue%5D=&columns%5B1%5D%5Bsearch%5D%5Bregex%5D=false&columns%5B2%5D%5Bdata%5D=isp&columns%5B2%5D%5Bname%5D=&columns%5B2%5D%5Bsearchable%5D=true&columns%5B2%5D%5Borderable%5D=true&columns%5B2%5D%5Bsearch%5D%5Bvalue%5D=&columns%5B2%5D%5Bsearch%5D%5Bregex%5D=false&columns%5B3%5D%5Bdata%5D=country&columns%5B3%5D%5Bname%5D=&columns%5B3%5D%5Bsearchable%5D=true&columns%5B3%5D%5Borderable%5D=true&columns%5B3%5D%5Bsearch%5D%5Bvalue%5D=&columns%5B3%5D%5Bsearch%5D%5Bregex%5D=false&columns%5B4%5D%5Bdata%5D=client&columns%5B4%5D%5Bname%5D=&columns%5B4%5D%5Bsearchable%5D=true&columns%5B4%5D%5Borderable%5D=true&columns%5B4%5D%5Bsearch%5D%5Bvalue%5D=&columns%5B4%5D%5Bsearch%5D%5Bregex%5D=false&columns%5B5%5D%5Bdata%5D=clientVersion&columns%5B5%5D%5Bname%5D=&columns%5B5%5D%5Bsearchable%5D=true&columns%5B5%5D%5Borderable%5D=true&columns%5B5%5D%5Bsearch%5D%5Bvalue%5D=&columns%5B5%5D%5Bsearch%5D%5Bregex%5D=false&columns%5B6%5D%5Bdata%5D=os&columns%5B6%5D%5Bname%5D=&columns%5B6%5D%5Bsearchable%5D=true&columns%5B6%5D%5Borderable%5D=true&columns%5B6%5D%5Bsearch%5D%5Bvalue%5D=&columns%5B6%5D%5Bsearch%5D%5Bregex%5D=false&columns%5B7%5D%5Bdata%5D=lastUpdate&columns%5B7%5D%5Bname%5D=&columns%5B7%5D%5Bsearchable%5D=true&columns%5B7%5D%5Borderable%5D=true&columns%5B7%5D%5Bsearch%5D%5Bvalue%5D=&columns%5B7%5D%5Bsearch%5D%5Bregex%5D=false&columns%5B8%5D%5Bdata%5D=inSync&columns%5B8%5D%5Bname%5D=&columns%5B8%5D%5Bsearchable%5D=true&columns%5B8%5D%5Borderable%5D=true&columns%5B8%5D%5Bsearch%5D%5Bvalue%5D=&columns%5B8%5D%5Bsearch%5D%5Bregex%5D=false&order%5B0%5D%5Bcolumn%5D=0&order%5B0%5D%5Bdir%5D=asc&search%5Bvalue%5D=&search%5Bregex%5D=false&length=100&start={offset}"
    res = hr(url, proxy=proxy, user_agent=CHROME_USER_AGENT)
    if res.is_error():
        return res.to_err_result()
    try:
        return res.to_ok_result(SearchResult(**res.json))
    except Exception as e:
        return res.to_err_result(f"exception: {e}")



import importlib.metadata
import sys
from typing import NoReturn

from loguru import logger
from mm_std import Err
from rich.live import Live
from rich.table import Table

from mm_eth import erc20, rpc
from mm_eth.cli import calcs
from mm_eth.utils import from_wei_str


def get_nonce(nodes: list[str] | str, address: str, log_prefix: str | None = None) -> int | None:
    res = rpc.eth_get_transaction_count(nodes, address, attempts=5)
    prefix = log_prefix or address
    logger.debug(f"{prefix}: nonce={res.ok_or_err()}")
    if isinstance(res, Err):
        logger.info(f"{prefix}: nonce error, {res.err}")
        return None
    return res.ok


def get_base_fee(nodes: list[str], log_prefix: str | None = None) -> int | None:
    res = rpc.get_base_fee_per_gas(nodes)
    prefix = _get_prefix(log_prefix)
    logger.debug(f"{prefix}base_fee={res.ok_or_err()}")
    if isinstance(res, Err):
        logger.info(f"{prefix}base_fee error, {res.err}")
        return None
    return res.ok


def calc_max_fee_per_gas(nodes: list[str], max_fee_per_gas: str, log_prefix: str | None = None) -> int | None:
    if "base" in max_fee_per_gas.lower():
        base_fee = get_base_fee(nodes, log_prefix)
        if base_fee is None:
            return None
        return calcs.calc_var_wei_value(max_fee_per_gas, var_name="base", var_value=base_fee)
    return calcs.calc_var_wei_value(max_fee_per_gas)


def is_max_fee_per_gas_limit_exceeded(
    max_fee_per_gas: int,
    max_fee_per_gas_limit: str | None,
    log_prefix: str | None = None,
) -> bool:
    if max_fee_per_gas_limit is None:
        return False
    max_fee_per_gas_limit_value = calcs.calc_var_wei_value(max_fee_per_gas_limit)
    if max_fee_per_gas > max_fee_per_gas_limit_value:
        prefix = _get_prefix(log_prefix)
        logger.info(
            "{}max_fee_per_gas_limit is exeeded, max_fee_per_gas={}, max_fee_per_gas_limit={}",
            prefix,
            from_wei_str(max_fee_per_gas, "gwei"),
            from_wei_str(max_fee_per_gas_limit_value, "gwei"),
        )
        return True
    return False


def is_value_less_min_limit(
    value_min_limit: str | None,
    value: int,
    value_unit: str,
    decimals: int | None = None,
    log_prefix: str | None = None,
) -> bool:
    if value_min_limit is None:
        return False
    if value < calcs.calc_var_wei_value(value_min_limit, decimals=decimals):
        prefix = _get_prefix(log_prefix)
        logger.info("{}value is less min limit, value={}", prefix, from_wei_str(value, value_unit, decimals=decimals))
        return True
    return False


def calc_gas(
    *,
    nodes: list[str],
    gas: str,
    from_address: str,
    to_address: str,
    value: int | None = None,
    data: str | None = None,
    log_prefix: str | None = None,
) -> int | None:
    estimate_value = None
    if "estimate" in gas.lower():
        prefix = _get_prefix(log_prefix)
        res = rpc.eth_estimate_gas(nodes, from_address, to_address, data=data, value=value, attempts=5)
        logger.debug(f"{prefix}gas_estimate={res.ok_or_err()}")
        if isinstance(res, Err):
            logger.info(f"{prefix}estimate_gas error, {res.err}")
            return None
        estimate_value = res.ok
    return calcs.calc_var_wei_value(gas, var_name="estimate", var_value=estimate_value)


def calc_eth_value(
    *,
    nodes: list[str],
    value_str: str,
    address: str,
    gas: int | None = None,
    max_fee_per_gas: int | None = None,
    log_prefix: str | None = None,
) -> int | None:
    balance_value = None
    if "balance" in value_str.lower():
        prefix = _get_prefix(log_prefix)
        res = rpc.eth_get_balance(nodes, address, attempts=5)
        logger.debug(f"{prefix}balance={res.ok_or_err()}")
        if isinstance(res, Err):
            logger.info(f"{prefix}balance error, {res.err}")
            return None
        balance_value = res.ok
    value = calcs.calc_var_wei_value(value_str, var_name="balance", var_value=balance_value)
    if "balance" in value_str.lower() and gas is not None and max_fee_per_gas is not None:
        value = value - gas * max_fee_per_gas
    return value


def calc_erc20_value(
    *,
    nodes: list[str],
    value_str: str,
    wallet_address: str,
    token_address: str,
    decimals: int,
    log_prefix: str | None = None,
) -> int | None:
    value_str = value_str.lower()
    balance_value = None
    if "balance" in value_str:
        prefix = _get_prefix(log_prefix)
        res = erc20.get_balance(nodes, token_address, wallet_address, attempts=5)
        logger.debug(f"{prefix}balance={res.ok_or_err()}")
        if isinstance(res, Err):
            logger.info(f"{prefix}balance error, {res.err}")
            return None
        balance_value = res.ok
    return calcs.calc_var_wei_value(value_str, var_name="balance", var_value=balance_value, decimals=decimals)


def print_balances(
    rpc_nodes: list[str],
    addresses: list[str],
    *,
    token_address: str | None = None,
    token_decimals: int | None = None,
    round_ndigits: int = 5,
) -> None:
    table = Table(title="balances")
    table.add_column("n")
    table.add_column("address")
    table.add_column("nonce")
    table.add_column("balance, eth")
    if token_address is not None and token_decimals is not None:
        table.add_column("token, t")
    with Live(table, refresh_per_second=0.5):
        count = 0
        for address in addresses:
            count += 1
            nonce = str(rpc.eth_get_transaction_count(rpc_nodes, address, attempts=5).ok_or_err())
            balance = rpc.eth_get_balance(rpc_nodes, address, attempts=5).map_or_else(
                lambda err: err,
                lambda ok: from_wei_str(ok, "eth", round_ndigits),
            )
            row: list[str] = [str(count), address, nonce, balance]
            if token_address is not None and token_decimals is not None:
                erc20_balance = erc20.get_balance(rpc_nodes, token_address, address, attempts=5).map_or_else(
                    lambda err: err,
                    lambda ok: from_wei_str(ok, "t", decimals=token_decimals, round_ndigits=round_ndigits),
                )
                row.append(erc20_balance)
            table.add_row(*row)


def fatal(message: str) -> NoReturn:
    print(f"error: {message}", file=sys.stderr)  # noqa: T201
    sys.exit(1)


def get_version() -> str:
    return importlib.metadata.version("mm-eth")


def _get_prefix(log_prefix: str | None) -> str:
    prefix = log_prefix or ""
    if prefix:
        prefix += ": "
    return prefix



from __future__ import annotations

import time
from subprocess import Popen  # nosec

from mm_std import Err, Ok, Result
from mm_std.net import get_free_local_port

from mm_eth import account, rpc


class Anvil:
    def __init__(self, *, chain_id: int, port: int, mnemonic: str):
        self.chain_id = chain_id
        self.port = port
        self.mnemonic = mnemonic
        self.process: Popen | None = None  # type: ignore[type-arg]

    def start_process(self) -> None:
        cmd = f"anvil -m '{self.mnemonic}' -p {self.port} --chain-id {self.chain_id}"
        self.process = Popen(cmd, shell=True)  # nosec
        time.sleep(3)

    def stop(self) -> None:
        if self.process:
            self.process.kill()

    def check(self) -> bool:
        res = rpc.eth_chain_id(self.rpc_url)
        return isinstance(res, Ok) and res.ok == self.chain_id

    @property
    def rpc_url(self) -> str:
        return f"http://localhost:{self.port}"

    @classmethod
    def launch(
        cls,
        chain_id: int = 31337,
        port: int | None = None,
        mnemonic: str = "",
        attempts: int = 3,
    ) -> Result[Anvil]:
        if not mnemonic:
            mnemonic = account.generate_mnemonic()

        for _ in range(attempts):
            if not port:
                port = get_free_local_port()
            anvil = Anvil(chain_id=chain_id, port=port, mnemonic=mnemonic)
            anvil.start_process()
            if anvil.check():
                return Ok(anvil)
            port = get_free_local_port()

        return Err("can't lauch anvil")



from mm_std import Ok, PrintFormat, print_json, print_plain
from rich.live import Live
from rich.table import Table

from mm_eth import rpc
from mm_eth.utils import from_wei_str, name_network


def run(urls: list[str], print_format: PrintFormat, proxy: str | None) -> None:
    json_result: dict[str, object] = {}
    table = Table(title="nodes")
    if print_format == PrintFormat.TABLE:
        table.add_column("url")
        table.add_column("chain_id")
        table.add_column("chain_name")
        table.add_column("block_number")
        table.add_column("base_fee")

    with Live(table, refresh_per_second=0.5):
        for url in urls:
            chain_id_res = rpc.eth_chain_id(url, timeout=10, proxies=proxy)
            chain_id = chain_id_res.ok_or_err()
            chain_name = ""
            if isinstance(chain_id_res, Ok):
                chain_name = name_network(chain_id_res.ok)
            block_number = rpc.eth_block_number(url, timeout=10, proxies=proxy).ok_or_err()
            base_fee = rpc.get_base_fee_per_gas(url, timeout=10, proxies=proxy).map_or_else(
                lambda err: err,
                lambda ok: from_wei_str(ok, "gwei"),
            )

            json_result[url] = {
                "chain_id": chain_id,
                "chain_name": chain_name,
                "block_number": block_number,
                "base_fee": base_fee,
            }
            if print_format == PrintFormat.TABLE:
                table.add_row(url, str(chain_id), chain_name, str(block_number), base_fee)
            print_plain(f"url: {url}", print_format)
            print_plain(f"chain_id: {chain_id}", print_format)
            print_plain(f"chain_name: {chain_name}", print_format)
            print_plain(f"block_number: {block_number}", print_format)
            print_plain(f"base_fee: {base_fee}", print_format)
            print_plain("", print_format)

    print_json(json_result, print_format)



import re
from decimal import Decimal, localcontext
from typing import Any, cast

import eth_utils
import pydash
from eth_typing import HexStr
from hexbytes import HexBytes
from mm_std import Err, Ok, Result, number_with_separator, random_choice
from mm_std.random_ import random_str_choice
from pydantic import BaseModel
from web3 import Web3
from web3.types import Wei

from mm_eth.types import Nodes, Proxies


def parse_addresses(data: str) -> list[str]:
    data = data.lower()
    result = []
    for word in data.split():
        if len(word) == 42 and re.match("0x[a-f0-9]{40}", word):
            result.append(word)
    return pydash.uniq(result)


def to_token_wei(value: str | int, decimals: int) -> int:
    if isinstance(value, int):
        return value
    elif isinstance(value, str):
        value = value.lower().replace(" ", "").strip()
        if value.endswith("t"):
            value = value.replace("t", "")
            return int(Decimal(value) * 10**decimals)
        elif value.isdigit():
            return int(value)

    raise ValueError("wrong value" + value)


def to_wei(value: str | int | Decimal, decimals: int | None = None) -> Wei:
    if isinstance(value, int):
        return Wei(value)
    elif isinstance(value, Decimal):
        if value != value.to_integral_value():
            raise ValueError(f"value must be integral number: {value}")
        return Wei(int(value))
    elif isinstance(value, str):
        value = value.lower().replace(" ", "").strip()
        if value.endswith("navax"):  # https://snowtrace.io/unitconverter
            value = value.replace("navax", "")
            return Wei(int(Decimal(value) * 10**9))
        if value.endswith("gwei"):
            value = value.replace("gwei", "")
            return Wei(int(Decimal(value) * 1000000000))
        if value.endswith("ether"):
            value = value.replace("ether", "")
            return Wei(int(Decimal(value) * 1000000000000000000))
        if value.endswith("eth"):
            value = value.replace("eth", "")
            return Wei(int(Decimal(value) * 1000000000000000000))
        if value.endswith("t"):
            if decimals is None:
                raise ValueError("t without decimals")
            value = value.removesuffix("t")
            return Wei(int(Decimal(value) * 10**decimals))
        if value.isdigit():
            return Wei(int(value))
        raise ValueError("wrong value " + value)

    else:
        raise ValueError(f"value has a wrong type: {type(value)}")


def from_wei(
    value: int,
    unit: str,
    round_ndigits: int | None = None,
    decimals: int | None = None,
) -> Decimal:
    if value == 0:
        return Decimal(0)

    is_negative = value < 0
    if unit.lower() == "eth":
        unit = "ether"

    if unit.lower() == "t":
        if decimals is None:
            raise ValueError("t without decimals")
        with localcontext() as ctx:
            ctx.prec = 999
            res = Decimal(value=abs(value), context=ctx) / Decimal(10**decimals)
    else:
        res = cast(Decimal, eth_utils.from_wei(abs(value), unit))
    if round_ndigits is not None:
        res = round(res, ndigits=round_ndigits)
    return -1 * res if is_negative else res


def from_wei_str(
    value: int,
    unit: str,
    round_ndigits: int | None = None,
    print_unit_name: bool = True,
    decimals: int | None = None,
) -> str:
    res = format(from_wei(value, unit, round_ndigits, decimals=decimals), "f")
    if unit == "ether":
        unit = "eth"
    if print_unit_name:
        res += unit
    return res


def from_token_wei_str(value: int, decimals: int, symbol: str = "", round_ndigits: int | None = None) -> str:
    res = value / 10**decimals
    if round_ndigits is not None:
        res = round(res, ndigits=round_ndigits)
    if symbol:
        res = f"{res} {symbol}"
    return str(res)


def to_wei_token(value: str | int | Decimal, symbol: str, decimals: int) -> int:
    if isinstance(value, int):
        return value
    elif isinstance(value, Decimal):
        if value != value.to_integral_value():
            raise ValueError(f"value must be integral number: {value}")
        return int(value)
    elif isinstance(value, str):
        value = value.lower().replace(" ", "").strip()
        if value.isdigit():
            return int(value)
        try:
            return int(Decimal(value.replace(symbol.lower(), "").strip()) * (10**decimals))
        except Exception as e:
            raise ValueError from e
    else:
        raise ValueError(f"value has a wrong type: {type(value)}")


def to_checksum_address(address: str) -> str:
    return Web3.to_checksum_address(address)


def hex_to_bytes(data: str) -> bytes:
    return eth_utils.to_bytes(hexstr=HexStr(data))


def get_chain_name(chain_id: int | str) -> str:
    chain_id = str(chain_id)
    if chain_id == "1":
        return "mainnet"
    if chain_id == "3":
        return "ropsten"
    if chain_id == "5":
        return "goerli"
    return chain_id


def to_human_readable_tx(tx: dict[str, Any] | BaseModel) -> dict[str, object]:
    if isinstance(tx, BaseModel):
        tx = tx.model_dump()
    tx["human_readable"] = {}
    tx["human_readable"]["gas_price"] = str(tx["gas_price"] / 10**9) + " gwei"
    tx["human_readable"]["value"] = str(tx["value"] / 10**18) + " ether"
    tx["human_readable"]["gas"] = number_with_separator(tx["gas"])
    if tx.get("chain_id") is not None:
        tx["human_readable"]["chain_id"] = get_chain_name(tx["chain_id"])

    return tx


def truncate_hex_str(hex_str: str, digits: int = 4, replace_str: str = "...") -> str:
    if not hex_str.startswith("0x") and not hex_str.startswith("0X"):
        raise ValueError("truncate_hex_str: hex_str must start with 0x")
    if digits <= 0:
        raise ValueError("truncate_hex_str: digits must be more than zero")
    hex_str = hex_str.removeprefix("0x").removeprefix("0X")
    if digits * 2 >= len(hex_str):
        raise ValueError("truncate_hex_str: digits is too large")
    return "0x" + hex_str[:digits] + replace_str + hex_str[-1 * digits :]


def log_topic_to_address(topic: HexBytes | str) -> str:
    if isinstance(topic, HexBytes):
        result = topic.hex()[-40:]
    else:
        result = topic[-40:]
    if not result.startswith("0x"):
        result = f"0x{result}"
    return result


def get_w3(rpc_url: str, timeout: int | None = None, proxy: str | None = None) -> Web3:
    request_kwargs: dict[str, object] = {"timeout": timeout}
    if proxy:
        request_kwargs["proxies"] = {"http": proxy, "https": proxy}
    return Web3(Web3.HTTPProvider(rpc_url, request_kwargs=request_kwargs))


def name_network(chain_id: int) -> str:
    match chain_id:
        case 1:
            return "Ethereum Mainnet"
        case 5:
            return "Goerli"
        case 10:
            return "OP Mainnet"
        case 280:
            return "zkSync Era Testnet"
        case 324:
            return "zkSync Era Mainnet"
        case 420:
            return "Optimism Goerli Testnet"
        case 42161:
            return "Arbitrum One"
        case 43113:
            return "Avalanche Fuji Testnet"
        case 43114:
            return "Avalanche C-Chain"
        case 421613:
            return "Arbitrum Goerli"
        case _:
            return ""


def hex_str_to_int(value: str) -> Result[int]:
    try:
        return Ok(int(value, 16))
    except Exception:
        return Err(f"can't convert to int: {value}")


def random_node(nodes: Nodes, remove_slash: bool = True) -> str:
    node = cast(str, random_choice(nodes))
    if remove_slash and node.endswith("/"):
        node = node.removesuffix("/")
    return node


def random_proxy(proxies: Proxies) -> str | None:
    return random_str_choice(proxies)



import os
from decimal import Decimal, InvalidOperation
from pathlib import Path

from mm_std import str_to_list

from mm_eth.cli import calcs
from mm_eth.utils import to_wei


def wei_validator(v: str | None) -> int | None:
    if v is None:
        return None
    return to_wei(v)


def log_validator(v: str | None) -> str | None:
    if v is None:
        return None
    log_file = Path(v).expanduser()
    log_file.touch(exist_ok=True)
    if not (log_file.is_file() and os.access(log_file, os.W_OK)):
        raise ValueError(f"wrong log path: {v}")
    return v


def nodes_validator(v: str | list[str] | None) -> list[str]:
    if v is None:
        return []
    if isinstance(v, str):
        return str_to_list(v, unique=True, remove_comments=True, split_line=True)
    return v


def addresses_validator(v: str | list[str] | None) -> list[str]:
    if v is None:
        return []
    if isinstance(v, str):
        return str_to_list(v, unique=True, remove_comments=True, split_line=True, lower=True)
    return v


def delay_validator(v: str | Decimal) -> Decimal | tuple[Decimal, Decimal]:
    if isinstance(v, int | float):
        return Decimal(str(v))
    elif isinstance(v, str):
        arr = [a.strip() for a in v.split("-")]
        if len(arr) != 2:
            raise ValueError("wrong delay value")
        try:
            return Decimal(arr[0]), Decimal(arr[1])
        except InvalidOperation:
            raise ValueError("wrong delay value") from None
    raise ValueError("wrong delay value")


def is_valid_calc_var_wei_value(value: str | None, base_name: str = "var", decimals: int | None = None) -> bool:
    if value is None:
        return True  # check for None on BaseModel.field type level
    try:
        calcs.calc_var_wei_value(value, var_value=123, var_name=base_name, decimals=decimals)
        return True
    except ValueError:
        return False


def is_valid_calc_decimal_value(value: str | None) -> bool:
    if value is None:
        return True  # check for None on BaseModel.field type level
    try:
        calcs.calc_decimal_value(value)
        return True
    except ValueError:
        return False


def is_valid_calc_function_args(value: str | None) -> bool:
    if value is None:
        return True
    try:
        calcs.calc_function_args(value)
        return True
    except ValueError:
        return False



import time
from pathlib import Path
from typing import Self

from loguru import logger
from mm_std import Err, Ok, fatal, print_json, str_to_list, utc_now
from pydantic import Field, StrictStr, field_validator, model_validator

from mm_eth import erc20, rpc
from mm_eth.account import create_private_keys_dict, private_to_address
from mm_eth.cli import calcs, cli_helpers, cli_utils, validators
from mm_eth.cli.cli_helpers import get_version
from mm_eth.cli.cli_utils import BaseConfig
from mm_eth.utils import from_wei_str


class Config(BaseConfig):
    class Tx(BaseConfig):
        from_address: str
        to_address: str

    token: str
    decimals: int
    nodes: list[StrictStr]
    chain_id: int
    private_keys: dict[str, str] = Field(default_factory=dict)
    private_keys_file: str | None = None
    max_fee_per_gas: str
    max_fee_per_gas_limit: str | None = None
    max_priority_fee_per_gas: str
    value: str
    value_min_limit: str | None = None
    gas: str
    addresses_map: str | None = None
    addresses_from_file: str | None = None
    addresses_to_file: str | None = None
    delay: str | None = None  # in seconds
    round_ndigits: int = 5
    log_debug: str | None = None
    log_info: str | None = None
    txs: list[Tx] = Field(default_factory=list)

    @property
    def from_addresses(self) -> list[str]:
        return [tx.from_address for tx in self.txs]

    @field_validator("log_debug", "log_info", mode="before")
    def log_validator(cls, v: str | None) -> str | None:
        return validators.log_validator(v)

    @field_validator("nodes", mode="before")
    def nodes_validator(cls, v: str | list[str] | None) -> list[str]:
        return validators.nodes_validator(v)

    @field_validator("private_keys", mode="before")
    def private_keys_validator(cls, v: str | list[str] | None) -> dict[str, str]:
        if v is None:
            return {}
        if isinstance(v, str):
            return create_private_keys_dict(str_to_list(v, unique=True, remove_comments=True))
        return create_private_keys_dict(v)

    # noinspection DuplicatedCode
    @model_validator(mode="after")
    def final_validator(self) -> Self:
        # load private keys from file
        if self.private_keys_file:
            file = Path(self.private_keys_file).expanduser()
            if not file.is_file():
                raise ValueError("can't read private_keys_file")
            for line in file.read_text().strip().split("\n"):
                line = line.strip()
                address = private_to_address(line)
                if address is None:
                    raise ValueError("there is not a private key in private_keys_file")
                self.private_keys[address.lower()] = line

        # max_fee_per_gas
        if not validators.is_valid_calc_var_wei_value(self.max_fee_per_gas, "base"):
            raise ValueError(f"wrong max_fee_per_gas: {self.max_fee_per_gas}")

        # max_fee_per_gas_limit
        if not validators.is_valid_calc_var_wei_value(self.max_fee_per_gas_limit, "base"):
            raise ValueError(f"wrong max_fee_per_gas_limit: {self.max_fee_per_gas_limit}")

        # max_priority_fee_per_gas
        if not validators.is_valid_calc_var_wei_value(self.max_priority_fee_per_gas):
            raise ValueError(f"wrong max_priority_fee_per_gas: {self.max_priority_fee_per_gas}")

        # value
        if not validators.is_valid_calc_var_wei_value(self.value, "balance", decimals=self.decimals):
            raise ValueError(f"wrong value: {self.value}")

        # value_min_limit
        if not validators.is_valid_calc_var_wei_value(self.value_min_limit, decimals=self.decimals):
            raise ValueError(f"wrong value_min_limit: {self.value_min_limit}")

        # gas
        if not validators.is_valid_calc_var_wei_value(self.gas, "estimate"):
            raise ValueError(f"wrong gas: {self.gas}")

        # delay
        if not validators.is_valid_calc_decimal_value(self.delay):
            raise ValueError(f"wrong delay: {self.delay}")

        # txs
        if self.addresses_map:
            for tx in cli_utils.load_tx_addresses_from_str(self.addresses_map):
                self.txs.append(Config.Tx(from_address=tx[0], to_address=tx[1]))
        if self.addresses_from_file and self.addresses_to_file:
            self.txs.extend(
                Config.Tx(from_address=tx[0], to_address=tx[1])
                for tx in cli_utils.load_tx_addresses_from_files(self.addresses_from_file, self.addresses_to_file)
            )
        if not self.txs:
            raise ValueError("txs is empty")

        return self


# noinspection DuplicatedCode
def run(
    config_path: str,
    *,
    print_balances: bool,
    print_config: bool,
    debug: bool,
    no_receipt: bool,
    emulate: bool,
) -> None:
    config = cli_utils.read_config(Config, config_path)
    if print_config:
        print_json(config.model_dump(exclude={"private_keys", "addresses_map"}))
        exit(0)

    cli_utils.init_logger(debug, config.log_debug, config.log_info)
    cli_utils.check_nodes_for_chain_id(config.nodes, config.chain_id)

    # check decimals
    res = erc20.get_decimals(config.nodes[0], config.token)
    if isinstance(res, Err):
        fatal(f"can't get token decimals: {res.err}")
    if res.ok != config.decimals:
        fatal(f"config.decimals is wrong: {config.decimals} != {res.ok}")

    if print_balances:
        cli_helpers.print_balances(
            config.nodes,
            config.from_addresses,
            token_address=config.token,
            token_decimals=config.decimals,
            round_ndigits=config.round_ndigits,
        )
        exit(0)

    return _run_transfers(config, no_receipt=no_receipt, emulate=emulate)


# noinspection DuplicatedCode
def _run_transfers(config: Config, *, no_receipt: bool, emulate: bool) -> None:
    logger.info(f"started at {utc_now()} UTC")
    logger.debug(f"config={config.model_dump(exclude={'private_keys', 'addresses_map'}) | {'version':get_version()}}")
    cli_utils.check_private_keys(config.from_addresses, config.private_keys)
    for i, tx in enumerate(config.txs):
        _transfer(from_address=tx.from_address, to_address=tx.to_address, config=config, no_receipt=no_receipt, emulate=emulate)
        if not emulate and config.delay is not None and i < len(config.txs) - 1:
            delay_value = calcs.calc_decimal_value(config.delay)
            logger.debug(f"delay {delay_value} seconds")
            time.sleep(float(delay_value))
    logger.info(f"finished at {utc_now()} UTC")


# noinspection DuplicatedCode
def _transfer(*, from_address: str, to_address: str, config: Config, no_receipt: bool, emulate: bool) -> None:
    log_prefix = f"{from_address}->{to_address}"
    # get nonce
    nonce = cli_helpers.get_nonce(config.nodes, from_address, log_prefix)
    if nonce is None:
        return

    # get max_fee_per_gas
    max_fee_per_gas = cli_helpers.calc_max_fee_per_gas(config.nodes, config.max_fee_per_gas, log_prefix)
    if max_fee_per_gas is None:
        return

    # check max_fee_per_gas_limit
    if cli_helpers.is_max_fee_per_gas_limit_exceeded(max_fee_per_gas, config.max_fee_per_gas_limit, log_prefix):
        return

    # get gas
    gas = cli_helpers.calc_gas(
        nodes=config.nodes,
        gas=config.gas,
        from_address=from_address,
        to_address=config.token,
        data=erc20.encode_transfer_input_data(to_address, 1234),
        log_prefix=log_prefix,
    )
    if gas is None:
        return

    # get value
    value = cli_helpers.calc_erc20_value(
        nodes=config.nodes,
        value_str=config.value,
        wallet_address=from_address,
        token_address=config.token,
        decimals=config.decimals,
        log_prefix=log_prefix,
    )
    if value is None:
        return

    # value_min_limit
    if cli_helpers.is_value_less_min_limit(
        config.value_min_limit,
        value,
        "t",
        decimals=config.decimals,
        log_prefix=log_prefix,
    ):
        return

    max_priority_fee_per_gas = calcs.calc_var_wei_value(config.max_priority_fee_per_gas)
    tx_params = {
        "nonce": nonce,
        "max_fee_per_gas": max_fee_per_gas,
        "max_priority_fee_per_gas": max_priority_fee_per_gas,
        "gas": gas,
        "value": value,
        "to": to_address,
        "chain_id": config.chain_id,
    }

    # emulate?
    if emulate:
        msg = f"{log_prefix}: emulate,"
        msg += f" value={from_wei_str(value, 't',decimals=config.decimals, round_ndigits=config.round_ndigits)},"
        msg += f" max_fee_per_gas={from_wei_str(max_fee_per_gas, 'gwei', config.round_ndigits)},"
        msg += f" max_priority_fee_per_gas={from_wei_str(max_priority_fee_per_gas, 'gwei', config.round_ndigits)},"
        msg += f" gas={gas}"
        logger.info(msg)
        return

    logger.debug(f"{log_prefix}: tx_params={tx_params}")
    signed_tx = erc20.sign_transfer_tx(
        nonce=nonce,
        max_fee_per_gas=max_fee_per_gas,
        max_priority_fee_per_gas=max_priority_fee_per_gas,
        gas_limit=gas,
        private_key=config.private_keys[from_address],
        chain_id=config.chain_id,
        value=value,
        token_address=config.token,
        recipient_address=to_address,
    )
    res = rpc.eth_send_raw_transaction(config.nodes, signed_tx.raw_tx, attempts=5)
    if isinstance(res, Err):
        logger.info(f"{log_prefix}: send_error: {res.err}")
        return
    tx_hash = res.ok

    if no_receipt:
        msg = f"{log_prefix}: tx_hash={tx_hash}, value={from_wei_str(value, 't',decimals=config.decimals, round_ndigits=config.round_ndigits)}"  # noqa
        logger.info(msg)
    else:
        logger.debug(f"{log_prefix}: tx_hash={tx_hash}, wait receipt")
        while True:
            receipt_res = rpc.get_tx_status(config.nodes, tx_hash)
            if isinstance(receipt_res, Ok):
                status = "OK" if receipt_res.ok == 1 else "FAIL"
                msg = f"{log_prefix}: tx_hash={tx_hash}, value={from_wei_str(value, 't', decimals=config.decimals, round_ndigits=config.round_ndigits)}, status={status}"  # noqa
                logger.info(msg)
                break
            time.sleep(1)



from mm_eth import abi


def test_encode_function_signature():
    res = abi.encode_function_signature("transfer(address,uint256)")
    assert res == "0xa9059cbb"


def test_get_function_abi(erc20_abi):
    res = abi.get_function_abi(erc20_abi, "transfer")
    assert res["name"] == "transfer"


def test_encode_function_input_by_abi(erc20_abi):
    res = abi.encode_function_input_by_abi(erc20_abi, "transfer", ["0x2D88bd70Eb6c20302D4cdD69abeBEea02deEBEAE", 123456])
    input_data = "0xa9059cbb0000000000000000000000002d88bd70eb6c20302d4cdd69abebeea02deebeae000000000000000000000000000000000000000000000000000000000001e240"  # noqa
    assert res == input_data


def test_encode_function_input_by_signature():
    res = abi.encode_function_input_by_signature(
        "transfer(address,uint256)",
        ["0x2D88bd70Eb6c20302D4cdD69abeBEea02deEBEAE", 123456],
    )
    input_data = "0xa9059cbb0000000000000000000000002d88bd70eb6c20302d4cdd69abebeea02deebeae000000000000000000000000000000000000000000000000000000000001e240"  # noqa
    assert res == input_data



import os.path
import random
import shutil
from dataclasses import dataclass
from pathlib import Path

from mm_std import Err, Ok, Result, run_command


@dataclass
class SolcResult:
    bin: str
    abi: str


def solc(contract_name: str, contract_path: str, tmp_dir: str) -> Result[SolcResult]:
    if tmp_dir.startswith("~"):
        tmp_dir = os.path.expanduser(tmp_dir)
    if contract_path.startswith("~"):
        contract_path = os.path.expanduser(contract_path)
    work_dir = f"{tmp_dir}/solc_{contract_name}_{random.randint(0, 100_000_000)}"
    abi_path = f"{work_dir}/{contract_name}.abi"
    bin_path = f"{work_dir}/{contract_name}.bin"
    try:
        Path(work_dir).mkdir(parents=True)
        cmd = f"solc -o '{work_dir}' --abi --bin --optimize {contract_path}"
        run_command(cmd)
        abi = Path(abi_path).read_text()
        bin_ = Path(bin_path).read_text()
        return Ok(SolcResult(bin=bin_, abi=abi))
    except Exception as e:
        return Err(f"exception: {e}")
    finally:
        shutil.rmtree(work_dir, ignore_errors=True)



import json

from mm_std import Err, PrintFormat, print_json, print_plain
from mm_std.fs import get_filename_without_extension

from mm_eth.cli import cli_helpers
from mm_eth.solc import solc


def run(contract_path: str, tmp_dir: str, print_format: PrintFormat) -> None:
    contract_name = get_filename_without_extension(contract_path)
    res = solc(contract_name, contract_path, tmp_dir)
    if isinstance(res, Err):
        cli_helpers.fatal(res.err)

    bin_ = res.ok.bin
    abi = res.ok.abi

    if print_format == PrintFormat.JSON:
        print_json({"bin": bin_, "abi": json.loads(abi)})
    else:
        print_plain("bin:")
        print_plain(bin_)
        print_plain("abi:")
        print_plain(abi)



import sys
import time
from pathlib import Path
from typing import TypeVar

import eth_utils
import yaml
from loguru import logger
from mm_std import Err, fatal, str_to_list, utc_now
from pydantic import BaseModel, ConfigDict, ValidationError
from rich.console import Console
from rich.table import Table

from mm_eth import account, rpc
from mm_eth.account import is_private_key
from mm_eth.cli import calcs


class BaseConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")


def check_nodes_for_chain_id(nodes: list[str], chain_id: int) -> None:
    for node in nodes:
        res = rpc.eth_chain_id(node, timeout=7)
        if isinstance(res, Err):
            fatal(f"can't get chain_id for {node}, error={res.err}")
        if res.ok != chain_id:
            fatal(f"node {node} has a wrong chain_id: {res.ok}")


def check_private_keys(addresses: list[str], private_keys: dict[str, str]) -> None:
    for address in addresses:
        address = address.lower()
        if address not in private_keys:
            fatal(f"no private key for {address}")
        if account.private_to_address(private_keys[address]) != address:
            fatal(f"no private key for {address}")


def delay(value: str | None) -> None:
    if value is None:
        return
    time.sleep(float(calcs.calc_decimal_value(value)))


T = TypeVar("T")


def read_config(config_cls: type[T], config_path: str) -> T:
    try:
        with open(config_path) as f:
            config = config_cls(**yaml.full_load(f))
            return config
    except ValidationError as err:
        table = Table(title="config validation errors")
        table.add_column("field")
        table.add_column("message")
        for e in err.errors():
            loc = e["loc"]
            field = str(loc[0]) if len(loc) > 0 else ""
            table.add_row(field, e["msg"])
        console = Console()
        console.print(table)
        exit(1)


def log(log_path: str | None, *messages: object) -> None:
    if log_path is None:
        return
    message = ", ".join([str(m) for m in messages])
    message = f"{utc_now()}, {message}\n"
    with open(Path(log_path).expanduser(), "a") as f:
        f.write(message)


def load_tx_addresses_from_str(v: str | None) -> list[tuple[str, str]]:
    result: list[tuple[str, str]] = []
    if v is None:
        return result
    for line in str_to_list(v, remove_comments=True):
        arr = line.split()
        if len(arr) == 2 and eth_utils.is_address(arr[0]) and eth_utils.is_address(arr[1]):
            result.append((arr[0].lower(), arr[1].lower()))
    return result


def load_tx_addresses_from_files(addresses_from_file: str, addresses_to_file: str) -> list[tuple[str, str]]:
    from_file = Path(addresses_from_file).expanduser()
    to_file = Path(addresses_to_file).expanduser()
    if not from_file.is_file():
        raise ValueError(f"can't read addresses from 'addresses_from_file={addresses_from_file}")
    if not to_file.is_file():
        raise ValueError(f"can't read addresses from 'addresses_to_file={addresses_to_file}")

    # get addresses_from
    addresses_from = []
    for line in from_file.read_text().strip().split("\n"):
        if not eth_utils.is_address(line):
            raise ValueError(f"illigal address in addresses_from_file: {line}")
        addresses_from.append(line.lower())

    # get addresses_to
    addresses_to = []
    for line in to_file.read_text().strip().split("\n"):
        if not eth_utils.is_address(line):
            raise ValueError(f"illigal address in addresses_to_file: {line}")
        addresses_to.append(line.lower())

    if len(addresses_from) != len(addresses_to):
        raise ValueError("len(addresses_from) != len(addresses_to)")

    return list(zip(addresses_from, addresses_to, strict=True))


def load_private_keys_from_file(private_keys_file: str) -> list[str]:
    result: list[str] = []
    for item in Path(private_keys_file).expanduser().read_text().split():
        if is_private_key(item):
            result.append(item)
    return result


def init_logger(debug: bool, log_debug_file: str | None, log_info_file: str | None) -> None:
    if debug:
        level = "DEBUG"
        format_ = "<green>{time:YYYY-MM-DD HH:mm:ss}</green> <level>{level}</level> {message}"
    else:
        level = "INFO"
        format_ = "{message}"

    logger.remove()
    logger.add(sys.stderr, format=format_, colorize=True, level=level)
    if log_debug_file:
        logger.add(Path(log_debug_file).expanduser(), format="{time:YYYY-MM-DD HH:mm:ss} {level} {message}")
    if log_info_file:
        logger.add(Path(log_info_file).expanduser(), format="{message}", level="INFO")


def public_rpc_url(url: str | None) -> str:
    if not url or url == "1":
        return "https://ethereum.publicnode.com"
    if url.startswith(("http://", "https://", "ws://", "wss://")):
        return url

    match url.lower():
        case "opbnb" | "204":
            return "https://opbnb-mainnet-rpc.bnbchain.org"
        case _:
            return url



from mm_std import Err, Ok, Result

from mm_eth.types import Nodes, Proxies
from mm_eth.utils import get_w3, random_node, random_proxy


def get_name(nodes: Nodes, address: str, timeout: int = 10, proxies: Proxies = None, attempts: int = 3) -> Result[str]:
    result = Err("not_started")
    for _ in range(attempts):
        try:
            w3 = get_w3(random_node(nodes), timeout=timeout, proxy=random_proxy(proxies))
            res = w3.ens.name(w3.to_checksum_address(address))  # type: ignore[union-attr]
            return Ok(res or "")
        except Exception as e:
            result = Err(e)
    return result



from pathlib import Path
from typing import Any

from mm_std import print_json

from mm_eth.account import generate_accounts, generate_mnemonic


def run(mnemonic: str, passphrase: str, limit: int, print_path: bool, path_prefix: str, save_file: str) -> None:  # nosec
    result: dict[str, Any] = {}
    if not mnemonic:
        mnemonic = generate_mnemonic()
    result["mnemonic"] = mnemonic
    if passphrase:
        result["passphrase"] = passphrase
    result["accounts"] = []
    for acc in generate_accounts(mnemonic=mnemonic, passphrase=passphrase, limit=limit, path_prefix=path_prefix):
        new_account = {"address": acc.address, "private": acc.private_key}
        if print_path:
            new_account["path"] = acc.path
        result["accounts"].append(new_account)
    print_json(result)

    if save_file:
        with open(Path(save_file).expanduser(), "w") as f:
            for account in result["accounts"]:
                f.write(f"{account['address']}\t{account['private']}\n")



import random
from decimal import Decimal

from mm_std.random_ import random_decimal
from mm_std.str import split_on_plus_minus_tokens

from mm_eth.utils import to_wei


def calc_var_wei_value(value: str, *, var_name: str = "var", var_value: int | None = None, decimals: int | None = None) -> int:
    if not isinstance(value, str):
        raise ValueError(f"value is not str: {value}")
    try:
        var_name = var_name.lower()
        result = 0
        for item in split_on_plus_minus_tokens(value.lower()):
            operator = item[0]
            item = item[1:]
            if item.isdigit():
                item_value = int(item)
            elif item.endswith("eth"):
                item = item.removesuffix("eth")
                item_value = int(Decimal(item) * 10**18)
            elif item.endswith("ether"):
                item = item.removesuffix("ether")
                item_value = int(Decimal(item) * 10**18)
            elif item.endswith("gwei"):
                item = item.removesuffix("gwei")
                item_value = int(Decimal(item) * 10**9)
            elif item.endswith("t"):
                if decimals is None:
                    raise ValueError("t without decimals")
                item = item.removesuffix("t")
                item_value = int(Decimal(item) * 10**decimals)
            elif item.endswith(var_name):
                if var_value is None:
                    raise ValueError("base value is not set")
                item = item.removesuffix(var_name)
                k = Decimal(item) if item else Decimal(1)
                item_value = int(k * var_value)
            elif item.startswith("random(") and item.endswith(")"):
                item = item.lstrip("random(").rstrip(")")
                arr = item.split(",")
                if len(arr) != 2:
                    raise ValueError(f"wrong value, random part: {value}")
                from_value = to_wei(arr[0], decimals=decimals)
                to_value = to_wei(arr[1], decimals=decimals)
                if from_value > to_value:
                    raise ValueError(f"wrong value, random part: {value}")
                item_value = random.randint(from_value, to_value)
            else:
                raise ValueError(f"wrong value: {value}")

            if operator == "+":
                result += item_value
            if operator == "-":
                result -= item_value

        return result
    except Exception as err:
        raise ValueError(f"wrong value: {value}, error={err}") from err


def calc_decimal_value(value: str) -> Decimal:
    value = value.lower().strip()
    if value.startswith("random(") and value.endswith(")"):
        arr = value.lstrip("random(").rstrip(")").split(",")
        if len(arr) != 2:
            raise ValueError(f"wrong value, random part: {value}")
        from_value = Decimal(arr[0])
        to_value = Decimal(arr[1])
        if from_value > to_value:
            raise ValueError(f"wrong value, random part: {value}")
        return random_decimal(from_value, to_value)
    return Decimal(value)


def calc_function_args(value: str) -> str:
    while True:
        if "random(" not in value:
            return value
        start_index = value.index("random(")
        stop_index = value.index(")", start_index)
        random_range = [int(v.strip()) for v in value[start_index + 7 : stop_index].split(",")]
        if len(random_range) != 2:
            raise ValueError("wrong random(from,to) template")
        rand_value = str(random.randint(random_range[0], random_range[1]))
        value = value[0:start_index] + rand_value + value[stop_index + 1 :]



from mm_std import fatal, print_plain

from mm_eth import account


def run(private_key: str) -> None:
    try:
        print_plain(account.private_to_address(private_key))
    except Exception as e:
        fatal(f"wrong private key: {e}")



from enum import Enum
from typing import Annotated, Union

import typer
from mm_std import PrintFormat, print_plain

from mm_eth.cli import cli_helpers
from mm_eth.cli.cmd import (
    balance_cmd,
    balances_cmd,
    call_contract_cmd,
    config_example_cmd,
    deploy_cmd,
    encode_input_data_cmd,
    mnemonic_cmd,
    node_cmd,
    private_key_cmd,
    rpc_cmd,
    send_contract_cmd,
    solc_cmd,
    token_cmd,
    transfer_erc20_cmd,
    transfer_eth_cmd,
    vault_cmd,
)

app = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=False, add_completion=False)


class ConfigExample(str, Enum):
    TRANSFER_ETH = "transfer-eth"
    TRANSFER_ERC20 = "transfer-erc20"
    BALANCES = "balances"
    CALL_CONTRACT = "call-contract"


@app.command(name="balance", help="Gen account balance")
def balance_command(
    wallet_address: Annotated[str, typer.Argument()],
    token_address: Annotated[Union[str | None], typer.Option("--token", "-t")] = None,  # noqa UP007
    rpc_url: Annotated[str, typer.Option("--url", "-u", envvar="ETH_RPC_URL")] = "",  # nosec
    wei: bool = typer.Option(False, "--wei", "-w", help="Print balances in wei units"),
    print_format: Annotated[PrintFormat, typer.Option("--format", "-f", help="Print format")] = PrintFormat.PLAIN,
) -> None:
    balance_cmd.run(rpc_url, wallet_address, token_address, wei, print_format)


@app.command(name="token", help="Get token info")
def token_command(
    token_address: Annotated[str, typer.Argument()],
    rpc_url: Annotated[str, typer.Option("--url", "-u", envvar="ETH_RPC_URL")] = "",
) -> None:
    token_cmd.run(rpc_url, token_address)


@app.command(name="node", help="Check RPC url")
def node_command(
    urls: Annotated[list[str], typer.Argument()],
    print_format: Annotated[PrintFormat, typer.Option("--format", "-f", help="Print format")] = PrintFormat.TABLE,
    proxy: Annotated[str, typer.Option("--proxy", "-p", help="Proxy")] = "",
) -> None:
    node_cmd.run(urls, print_format, proxy)


@app.command(name="mnemonic", help="Generate eth accounts based on a mnemonic")
def mnemonic_command(  # nosec
    mnemonic: Annotated[str, typer.Option("--mnemonic", "-m")] = "",
    passphrase: Annotated[str, typer.Option("--passphrase", "-pass")] = "",
    print_path: bool = typer.Option(False, "--print_path"),
    path_prefix: Annotated[str, typer.Option("--path")] = "m/44'/60'/0'/0",
    limit: int = typer.Option(10, "--limit", "-l"),
    save_file: str = typer.Option("", "--save", "-s", help="Save private keys to a file"),
) -> None:
    mnemonic_cmd.run(
        mnemonic,
        passphrase=passphrase,
        print_path=print_path,
        limit=limit,
        path_prefix=path_prefix,
        save_file=save_file,
    )


@app.command(name="private-key", help="Print an address for a private key")
def private_key_command(private_key: str) -> None:
    private_key_cmd.run(private_key)


@app.command(name="solc", help="Compile a solidity file")
def solc_command(
    contract_path: str,
    tmp_dir: str = "/tmp",  # nosec
    print_format: Annotated[PrintFormat, typer.Option("--format", "-f", help="Print format")] = PrintFormat.PLAIN,
) -> None:
    solc_cmd.run(contract_path, tmp_dir, print_format)


@app.command(name="vault", help="Save private keys to vault")
def vault_command(
    keys_url: Annotated[
        str,
        typer.Option(..., "--url", "-u", help="Url to keys, for example https://vault.site.com:8200/v1/kv/keys1"),
    ],
    vault_token: Annotated[
        str,
        typer.Option(..., "--token", "-t", prompt=True, hide_input=True, prompt_required=False, help="A vault token"),
    ],
    keys_file: Annotated[str, typer.Option(..., "--file", "-f", help="Path to a file with private keys")],
) -> None:
    vault_cmd.run(keys_url, vault_token, keys_file)


@app.command(name="rpc", help="Call a JSON-RPC method")
def rpc_command(
    method: Annotated[str, typer.Argument()] = "",
    params: Annotated[str, typer.Argument()] = "[]",
    rpc_url: Annotated[str, typer.Option("--url", "-u", envvar="ETH_RPC_URL", help="RPC node url")] = "",
    hex2dec: Annotated[bool, typer.Option("--hex2dec", "-d", help="Print result in decimal value")] = False,
) -> None:
    rpc_cmd.run(rpc_url, method, params, hex2dec)


@app.command(name="transfer-eth", help="Transfer eth / base token from one or many accounts")
def transfer_eth_command(
    config_path: str,
    print_balances: bool = typer.Option(False, "--balances", "-b", help="Print balances and exit"),
    print_config: bool = typer.Option(False, "--config", "-c", help="Print config and exit"),
    emulate: bool = typer.Option(False, "--emulate", "-e", help="Emulate transaction posting"),
    no_receipt: bool = typer.Option(False, "--no-receipt", "-nr", help="Don't wait for a tx receipt"),
    debug: bool = typer.Option(False, "--debug", "-d", help="Print debug info"),
) -> None:
    transfer_eth_cmd.run(
        config_path,
        print_balances=print_balances,
        print_config=print_config,
        debug=debug,
        no_receipt=no_receipt,
        emulate=emulate,
    )


@app.command(name="transfer-erc20", help="Transfer ERC20 token from one or many accounts")
def transfer_erc20_command(
    config_path: str,
    print_balances: bool = typer.Option(False, "--balances", "-b", help="Print balances and exit"),
    print_config: bool = typer.Option(False, "--config", "-c", help="Print config and exit"),
    emulate: bool = typer.Option(False, "--emulate", "-e", help="Emulate transaction posting"),
    no_receipt: bool = typer.Option(False, "--no-receipt", "-nr", help="Don't wait for a tx receipt"),
    debug: bool = typer.Option(False, "--debug", "-d", help="Print debug info"),
) -> None:
    transfer_erc20_cmd.run(
        config_path,
        print_balances=print_balances,
        print_config=print_config,
        debug=debug,
        no_receipt=no_receipt,
        emulate=emulate,
    )


@app.command(name="send-contract", help="Send transactions to a contract")
def send_contract_command(
    config_path: str,
    print_balances: bool = typer.Option(False, "--balances", "-b", help="Print balances and exit"),
    print_config: bool = typer.Option(False, "--config", "-c", help="Print config and exit"),
    emulate: bool = typer.Option(False, "--emulate", "-e", help="Emulate transaction posting"),
    no_receipt: bool = typer.Option(False, "--no-receipt", "-nr", help="Don't wait for a tx receipt"),
    debug: bool = typer.Option(False, "--debug", "-d", help="Print debug info"),
) -> None:
    send_contract_cmd.run(
        config_path,
        print_balances=print_balances,
        print_config=print_config,
        debug=debug,
        no_receipt=no_receipt,
        emulate=emulate,
    )


@app.command(name="balances", help="Print base and ERC20 token balances")
def balances_command(
    config_path: str,
    print_config: bool = typer.Option(False, "--config", "-c", help="Print config only and exit"),
    nonce: bool = typer.Option(False, "--nonce", "-n", help="Print nonce also"),
    wei: bool = typer.Option(False, "--wei", "-w", help="Show balances in WEI"),
) -> None:
    balances_cmd.run(config_path, print_config, wei, nonce)


@app.command(name="call-contract", help="Call a method on a contract")
def call_contract_command(
    config_path: str,
    print_config: bool = typer.Option(False, "--config", "-c", help="Print config only and exit"),
) -> None:
    call_contract_cmd.run(config_path, print_config)


@app.command(name="deploy", help="Deploy a smart contract onchain")
def deploy_command(
    config_path: str,
    print_config: bool = typer.Option(False, "--config", "-c", help="Print config only and exit"),
) -> None:
    deploy_cmd.run(config_path, print_config=print_config)


@app.command(name="config-example", help="Print an example of config for a command")
def config_example_command(command: Annotated[ConfigExample, typer.Argument()]) -> None:
    config_example_cmd.run(command)


@app.command(name="encode-input-data", help="Encode input data by a function signature")
def encode_input_data(
    funtion_signature: str = typer.Argument(help="Function signature, for example: transfer(address, uint256)"),
    args_str: str = typer.Argument(
        help="""Function arguments, as an array string. For example: '["0xA659FB44eB5d4bFaC1074Cb426b1b11D58D28308", 123]' """,
    ),
) -> None:
    encode_input_data_cmd.run(funtion_signature, args_str)


def version_callback(value: bool) -> None:
    if value:
        print_plain(f"mm-eth version: {cli_helpers.get_version()}")
        raise typer.Exit()


@app.callback()
def main(_version: bool = typer.Option(None, "--version", callback=version_callback, is_eager=True)) -> None:
    pass


if __name__ == "__main_":
    app()



from dataclasses import dataclass

from mm_std import Err, Ok, fatal, json_dumps
from pydantic import BaseModel, Field, field_validator
from rich.console import Console
from rich.live import Live
from rich.table import Table

from mm_eth import erc20, rpc
from mm_eth.cli import cli_utils, validators
from mm_eth.utils import from_token_wei_str, from_wei_str


class Config(BaseModel):
    addresses: list[str]
    tokens: list[str] = Field(default_factory=list)
    nodes: list[str]
    round_ndigits: int = 5

    @field_validator("nodes", mode="before")
    def nodes_validator(cls, v: str | list[str] | None) -> list[str]:
        return validators.nodes_validator(v)

    @field_validator("tokens", "addresses", mode="before")
    def addresses_validator(cls, v: str | list[str] | None) -> list[str]:
        return validators.addresses_validator(v)


@dataclass
class Token:
    address: str
    decimals: int
    symbol: str


def run(config_path: str, print_config: bool, wei: bool, show_nonce: bool) -> None:
    config = cli_utils.read_config(Config, config_path)
    if print_config:
        console = Console()
        console.print_json(json_dumps(config.model_dump()))
        exit(0)

    tokens = _get_tokens_info(config)

    table = Table(title="balances")
    table.add_column("address")
    if show_nonce:
        table.add_column("nonce")
    table.add_column("wei" if wei else "eth")
    for t in tokens:
        table.add_column(t.symbol)

    base_sum = 0
    token_sum: dict[str, int] = {t.address: 0 for t in tokens}
    with Live(table, refresh_per_second=0.5):
        for address in config.addresses:
            row = [address]
            if show_nonce:
                row.append(str(rpc.eth_get_transaction_count(config.nodes, address, attempts=5).ok_or_err()))

            base_balance_res = rpc.eth_get_balance(config.nodes, address, attempts=5)
            if isinstance(base_balance_res, Ok):
                base_sum += base_balance_res.ok
                if wei:
                    row.append(str(base_balance_res.ok))
                else:
                    row.append(
                        from_wei_str(base_balance_res.ok, "eth", round_ndigits=config.round_ndigits, print_unit_name=False),
                    )
            else:
                row.append(base_balance_res.err)

            for t in tokens:
                token_balance_res = erc20.get_balance(config.nodes, t.address, address, attempts=5)
                if isinstance(token_balance_res, Ok):
                    token_sum[t.address] += token_balance_res.ok
                    if wei:
                        row.append(str(token_balance_res.ok))
                    else:
                        row.append(
                            from_token_wei_str(
                                token_balance_res.ok,
                                decimals=t.decimals,
                                round_ndigits=config.round_ndigits,
                            ),
                        )
                else:
                    row.append(token_balance_res.err)

            table.add_row(*row)

        sum_row = ["sum"]
        if show_nonce:
            sum_row.append("")
        if wei:
            sum_row.append(str(base_sum))
            for t in tokens:
                sum_row.append(str(token_sum[t.address]))
        else:
            sum_row.append(from_wei_str(base_sum, "eth", round_ndigits=config.round_ndigits, print_unit_name=False))
            for t in tokens:
                sum_row.append(from_token_wei_str(token_sum[t.address], t.decimals, round_ndigits=config.round_ndigits))
        table.add_row(*sum_row)


def _get_tokens_info(config: Config) -> list[Token]:
    result: list[Token] = []
    for address in config.tokens:
        decimals_res = erc20.get_decimals(config.nodes, address, attempts=5)
        if isinstance(decimals_res, Err):
            fatal(f"can't get token {address} decimals: {decimals_res.err}")
        decimal = decimals_res.ok

        symbols_res = erc20.get_symbol(config.nodes, address, attempts=5)
        if isinstance(symbols_res, Err):
            fatal(f"can't get token {address} symbol: {symbols_res.err}")
        symbol = symbols_res.ok

        result.append(Token(address=address, decimals=decimal, symbol=symbol))

    return result



import json
import pkgutil
from typing import cast

from eth_typing import ChecksumAddress, HexStr
from hexbytes import HexBytes
from mm_std import Err, Ok, Result
from pydantic import BaseModel, Field
from web3 import Web3
from web3.types import ABI, Nonce, TxParams, Wei

from mm_eth import abi, rpc
from mm_eth.rpc import rpc_call
from mm_eth.types import Nodes, Proxies
from mm_eth.utils import hex_str_to_int

L2_ETH_TOKEN_ADDRESS = Web3.to_checksum_address("0x000000000000000000000000000000000000800a")


class BridgeContracts(BaseModel):
    l1_erc20_Default_bridge: str = Field(..., alias="l1Erc20DefaultBridge")
    l2_erc20_Default_bridge: str = Field(..., alias="l2Erc20DefaultBridge")


def zks_get_bridge_contracts(
    rpc_urls: Nodes,
    timeout: int = 10,
    proxies: Proxies = None,
    attempts: int = 1,
) -> Result[BridgeContracts]:
    res = rpc_call(
        nodes=rpc_urls,
        method="zks_getBridgeContracts",
        params=[],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    )
    if isinstance(res, Err):
        return res

    try:
        return Ok(BridgeContracts(**res.ok), data=res.data)
    except Exception as e:
        return Err(f"exception: {e}", data=res.data)


def zks_get_main_contract(
    rpc_urls: Nodes,
    timeout: int = 10,
    proxies: Proxies = None,
    attempts: int = 1,
) -> Result[str]:
    return rpc_call(
        nodes=rpc_urls,
        method="zks_getMainContract",
        params=[],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    )


def zks_estimate_gas_l1_to_l2(
    rpc_urls: Nodes,
    timeout: int = 10,
    proxies: Proxies = None,
    attempts: int = 1,
) -> Result[int]:
    return rpc_call(
        nodes=rpc_urls,
        method="zks_estimateGasL1ToL2",
        params=[
            [
                {
                    "from": "0x1111111111111111111111111111111111111111",
                    "to": "0x2222222222222222222222222222222222222222",
                    "data": "0xffffffff",
                },
            ],
        ],
        timeout=timeout,
        proxies=proxies,
        attempts=attempts,
    ).and_then(hex_str_to_int)


def zksync_contract_abi() -> ABI:
    data = pkgutil.get_data(__name__, "abi/zksync.json")
    if data is None:
        raise RuntimeError("can't read abi/zksync.json")
    return cast(ABI, json.loads(data.decode()))


def get_l2_transactio_base_cost(
    eth_rpc: str,
    contract_address: ChecksumAddress,
    gas_price: Wei,
    l2_gas_limit: int = 10000000,
    l2_gas_per_pubdata_byte_limit: int = 800,
) -> Result[Wei]:
    w3 = Web3(Web3.HTTPProvider(eth_rpc))
    try:
        contract = w3.eth.contract(address=contract_address, abi=zksync_contract_abi())
        res = contract.functions.l2TransactionBaseCost(gas_price, l2_gas_limit, l2_gas_per_pubdata_byte_limit).call()
        return Ok(Wei(res))
    except Exception as e:
        return Err(str(e))


def deposit(
    *,
    eth_rpc: str,
    contract_address: ChecksumAddress,
    wallet_address: ChecksumAddress,
    private_key: str,
    gas_price: Wei,
    value: Wei,
    l2_gas_limit: int = 1_300_000,
    l2_gas_per_pubdata_byte_limit: int = 800,
    gas: int = 150096,
    nonce: Nonce | None = None,
) -> Result[str]:
    w3 = Web3(Web3.HTTPProvider(eth_rpc))
    try:
        # get nonce
        if nonce is None:
            res_nonce = rpc.eth_get_transaction_count(eth_rpc, wallet_address)
            if isinstance(res_nonce, Err):
                return res_nonce

            nonce = Nonce(res_nonce.ok)

        # get base_cost
        res = get_l2_transactio_base_cost(eth_rpc, contract_address, gas_price, l2_gas_limit, l2_gas_per_pubdata_byte_limit)
        if isinstance(res, Err):
            return res
        base_cost = Wei(res.ok)

        contract_address_l2 = wallet_address
        l2_value = value
        calldata = b""
        factory_deps = []  # type: ignore[var-annotated]
        refund_recipient = wallet_address

        contract = w3.eth.contract(address=contract_address, abi=zksync_contract_abi())
        function_call = contract.functions.requestL2Transaction(
            contract_address_l2,
            l2_value,
            calldata,
            l2_gas_limit,
            l2_gas_per_pubdata_byte_limit,
            factory_deps,
            refund_recipient,
        )

        transaction_data = function_call.build_transaction(
            {"from": wallet_address, "gas": gas, "gasPrice": gas_price, "nonce": nonce, "value": Wei(value + base_cost)},
        )
        signed_transaction = w3.eth.account.sign_transaction(transaction_data, private_key)
        txn = w3.eth.send_raw_transaction(signed_transaction.rawTransaction)
        return Ok(str(txn))
    except Exception as e:
        return Err(f"exception: {e}")


def withdraw(
    *,
    zksync_rpc: str,
    wallet_address: ChecksumAddress,
    private_key: str,
    value: Wei,
    chain_id: int,
) -> Result[HexBytes]:
    try:
        w3 = Web3(Web3.HTTPProvider(zksync_rpc))

        # get nonce
        nonce = w3.eth.get_transaction_count(wallet_address)

        input_data = abi.encode_function_signature("withdraw(address)") + abi.encode_data(
            ["address"],
            [wallet_address],
        ).removeprefix("0x")

        transaction_data: TxParams = {
            "from": wallet_address,
            "to": L2_ETH_TOKEN_ADDRESS,
            "nonce": nonce,
            "data": HexStr(input_data),
            "value": value,
            "chainId": chain_id,
        }
        gas = w3.eth.estimate_gas(transaction_data)
        transaction_data["gas"] = gas
        gas_price = w3.eth.gas_price
        transaction_data["gasPrice"] = gas_price
        signed_transaction = w3.eth.account.sign_transaction(transaction_data, private_key)
        txn = w3.eth.send_raw_transaction(signed_transaction.rawTransaction)
        return Ok(txn)

    except Exception as e:
        return Err(f"exception: {e}")



import pytest
from mm_eth import rpc, tx

pytestmark = pytest.mark.infura


def test_eth_block_number(infura):
    res = rpc.eth_block_number(infura())
    assert res.unwrap() > 9_000_000

    res = rpc.eth_block_number(infura(ws=True))
    assert res.unwrap() > 9_000_000


def test_eth_get_code(infura, address_bnb):
    res = rpc.eth_get_code(infura(), address_bnb)
    assert res.unwrap().startswith("0x606060405236")


def test_net_version(infura):
    res = rpc.net_version(infura())
    assert res.unwrap() == "1"


def test_eth_send_raw_transaction(infura, private_0, address_1):
    raw_tx = tx.sign_legacy_tx(nonce=0, gas_price=111, gas=21000, private_key=private_0, chain_id=1, value=222, to=address_1)
    res = rpc.eth_send_raw_transaction(infura(), raw_tx.raw_tx)
    assert res.unwrap_err().startswith("service_error: insufficient funds for")


def test_web3_client_version(infura):
    res = rpc.web3_client_version(infura())
    assert res.unwrap().startswith("Geth")


def test_net_peer_count(infura):
    res = rpc.net_peer_count(infura())
    assert res.unwrap() > 1


def test_eth_get_balance(infura, address_bnb):
    res = rpc.eth_get_balance(infura(), address_bnb)
    assert res.unwrap() > 1


def test_eth_get_block_by_number(infura):
    res = rpc.eth_get_block_by_number(infura(), 8972973, True)
    assert res.unwrap()["transactions"][0]["hash"] == "0x1bc1f41a0999c4ff4afe8f17704400ba0328b8b8bf60681fb809969c2127054a"


def test_eth_get_logs(infura, address_bnb):
    res = rpc.eth_get_logs(infura(), address=address_bnb, from_block=9_000_000, to_block=9_001_000)
    assert len(res.unwrap()) == 70


def test_eth_get_transaction_receipt(mainnet_archive_node):
    res = rpc.eth_get_transaction_receipt(
        mainnet_archive_node,
        "0xc10fc67499a037b6c2f14ae0c63b659b05bd7b553378202f96e777dd4843130f",
    )
    assert res.unwrap().block_number == 3154196


def test_eth_get_transaction_by_hash(mainnet_archive_node):
    res = rpc.eth_get_transaction_by_hash(
        mainnet_archive_node,
        "0x11f52d6cf97fd61c261f54d38134c1dcc32d32c4d60de6e64c31e776a46e6373",
    )
    assert res.unwrap().block_number == 9914284


def test_eth_call(infura, address_tether):
    # get tether balance of bnb address
    data = "0x27e235e3000000000000000000000000B8c77482e45F1F44dE1745F52C74426C631bDD52"
    res = rpc.eth_call(infura(), to=address_tether, data=data)
    assert int(res.unwrap(), 16) > 1_000_000


def test_eth_estimate_gas(infura, address_0, address_1):
    data = "0x27e235e30000000000000000000000003f5ce5fbfe3e9af3971dd833d26ba9b5c936f0be"
    res = rpc.eth_estimate_gas(infura(), from_=address_0, to=address_1, data=data)
    assert res.unwrap() == 21598


def test_eth_gas_price(anvil):
    res = rpc.eth_gas_price(anvil.rpc_url)
    assert res.unwrap() > 1_000_000


def test_eth_syncing(anvil):
    res = rpc.eth_syncing(anvil.rpc_url)
    assert res.unwrap() is False


def test_eth_chain_id(anvil):
    res = rpc.eth_chain_id(anvil.rpc_url)
    assert res.ok == anvil.chain_id


def test_get_base_fee_per_gas(anvil):
    res = rpc.get_base_fee_per_gas(anvil.rpc_url)
    assert res.unwrap() == 1000000000



import eth_utils
import rlp
from eth_utils import keccak

from mm_eth import abi


def get_deploy_contract_data(contract_bin: str, constructor_types: list[str], constructor_values: list[object]) -> str:
    constructor_data = ""
    if constructor_types and constructor_values:
        constructor_data = abi.encode_data(constructor_types, constructor_values)[2:]
    return contract_bin + constructor_data


def get_contract_address(sender_address: str, nonce: int) -> str:
    sender_bytes = eth_utils.to_bytes(hexstr=sender_address)
    raw = rlp.encode([sender_bytes, nonce])
    h = keccak(raw)
    address_bytes = h[12:]
    return eth_utils.to_checksum_address(address_bytes).lower()



from mm_eth import ethernodes
from mm_std import random_choice


def test_search_nodes(mm_proxies):
    res = ethernodes.search_nodes(offset=100, proxy=random_choice(mm_proxies))
    assert res.is_ok() and res.ok.records_total > 1000 and len(res.ok.data) == 100



from mm_std import Err, Ok, print_plain

from mm_eth import erc20, rpc
from mm_eth.cli.cli_utils import public_rpc_url


def run(rpc_url: str, token_address: str) -> None:
    rpc_url = public_rpc_url(rpc_url)
    name = erc20.get_name(rpc_url, token_address).ok_or_err()
    symbol = erc20.get_symbol(rpc_url, token_address).ok_or_err()
    decimals = erc20.get_decimals(rpc_url, token_address).ok_or_err()
    transfer_count = _calc_transfer_events(rpc_url, 100, token_address)

    print_plain(f"name: {name}")
    print_plain(f"symbol: {symbol}")
    print_plain(f"decimals: {decimals}")
    print_plain(f"transfer_count: {transfer_count}")


def _calc_transfer_events(rpc_url: str, last_block_limit: int, token_address: str) -> int | str:
    current_block_res = rpc.eth_block_number(rpc_url)
    if isinstance(current_block_res, Err):
        return current_block_res.err
    current_block = current_block_res.ok

    res = erc20.get_transfer_event_logs(rpc_url, token_address, current_block - last_block_limit, current_block)
    if isinstance(res, Ok):
        return len(res.ok)
    return res.err



import string
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, cast

import eth_abi
import eth_utils
import pydash
from eth_typing import HexStr
from pydantic import BaseModel
from web3 import Web3
from web3.auto import w3
from web3.types import ABI, ABIFunction

from mm_eth.utils import hex_to_bytes


@dataclass
class NameTypeValue:
    name: str
    type: str
    value: Any


class FunctionInput(BaseModel):
    function_abi: ABIFunction
    params: dict[str, Any]

    def decode_params_bytes(self) -> dict[str, Any]:
        result: dict[str, Any] = {}
        for k, v in self.params.items():
            if isinstance(v, bytes):
                try:
                    str_value = eth_utils.to_text(v)
                except UnicodeDecodeError:
                    str_value = eth_utils.to_hex(v)
                result[k] = "".join(filter(lambda x: x in string.printable, str_value))
            else:
                result[k] = v
        return result

    def function_signature(self) -> str:
        inputs = [i["name"] for i in self.function_abi["inputs"]]
        return self.function_abi["name"] + f"({','.join(inputs)})"

    def to_list(self, decode_bytes: bool = False) -> list[NameTypeValue]:
        result = []
        for param in self.function_abi["inputs"]:
            name = param["name"]
            type_ = param["type"]
            value = self.params[name]
            if decode_bytes and isinstance(value, bytes):
                try:
                    value = eth_utils.to_text(value)
                except UnicodeDecodeError:
                    value = eth_utils.to_hex(value)
            result.append(NameTypeValue(name, type_, value))
        return result


def decode_function_input(contract_abi: ABI, tx_input: str) -> FunctionInput:
    contract = w3.eth.contract(abi=contract_abi)
    func, params = contract.decode_function_input(HexStr(tx_input))
    return FunctionInput(function_abi=func.abi, params=params)


def get_function_abi(contr_abi: ABI, fn_name: str) -> ABIFunction:
    abi = pydash.find(contr_abi, lambda x: x.get("name", None) == fn_name and x.get("type", None) == "function")  # type:ignore
    if not abi:
        raise ValueError("can't find abi for function: " + fn_name)
    return cast(ABIFunction, abi)


def encode_function_input_by_abi(abi: ABI | ABIFunction, fn_name: str, args: list[Any]) -> HexStr:
    # if abi is contract_abi, get function_abi
    if isinstance(abi, Sequence):
        abi = get_function_abi(abi, fn_name)
    abi = cast(ABIFunction, abi)

    # need update all address values to checkSum version
    processed_args = []
    for idx, arg in enumerate(abi["inputs"]):
        if arg["type"] == "address":
            processed_args.append(eth_utils.to_checksum_address(args[idx]))
        else:
            processed_args.append(args[idx])

    return Web3().eth.contract(abi=[abi]).encodeABI(fn_name=fn_name, args=processed_args)  # type: ignore[no-any-return]


def encode_function_input_by_signature(func_signature: str, args: list[Any]) -> HexStr:
    if not func_signature.endswith(")"):
        raise ValueError(f"wrong func_signature={func_signature}. example: func1(uint256,address)")
    func_signature = func_signature.removesuffix(")")
    arr = func_signature.split("(")
    if len(arr) != 2:
        raise ValueError(f"wrong func_signature={func_signature}. example: func1(uint256,address)")
    func_name = arr[0]
    arg_types = [t.strip() for t in arr[1].split(",") if t.strip()]
    func_abi: ABIFunction = {
        "name": func_name,
        "type": "function",
        "inputs": [{"type": t} for t in arg_types],
    }
    return encode_function_input_by_abi(func_abi, func_name, args)


def encode_function_signature(func_name_with_types: str) -> HexStr:
    """input example 'transfer(address,uint256)'"""
    return HexStr(eth_utils.to_hex(Web3.keccak(text=func_name_with_types))[0:10])


def decode_data(types: list[str], data: str) -> tuple[Any, ...]:
    return eth_abi.decode(types, hex_to_bytes(data))


def encode_data(types: list[str], args: list[Any]) -> str:
    return eth_utils.to_hex(eth_abi.encode(types, args))


def parse_function_signatures(contract_abi: ABI) -> dict[str, str]:
    """returns dict, key: function_name_and_types, value: 4bytes signature"""
    result: dict[str, str] = {}
    for item in contract_abi:
        if item.get("type", None) == "function":
            function_name = item["name"]
            types = ",".join([i["type"] for i in item["inputs"]])
            function_name_and_types = f"{function_name}({types})"
            result[function_name_and_types] = encode_function_signature(function_name_and_types)
    return result



from mm_std import fatal, print_plain

from mm_eth import vault
from mm_eth.cli import cli_utils


def run(keys_url: str, vault_token: str, keys_file: str) -> None:
    private_keys = cli_utils.load_private_keys_from_file(keys_file)
    if not private_keys:
        fatal("private keys not found")

    res = vault.set_keys_from_vault(keys_url, vault_token, private_keys)
    if res.is_ok() and res.ok is True:
        print_plain(f"saved {len(private_keys)} private keys to the vault")
    else:
        fatal(f"error: {res.err}")



from dataclasses import dataclass

import eth_utils
from eth_account import Account
from eth_account.hdaccount import Mnemonic
from eth_keys import KeyAPI
from eth_typing import ChecksumAddress
from eth_utils import decode_hex

Account.enable_unaudited_hdwallet_features()  # type: ignore[no-untyped-call]

key_api = KeyAPI()


@dataclass
class NewAccount:
    path: str
    address: str
    private_key: str


def to_checksum_address(address: str) -> ChecksumAddress:
    return eth_utils.to_checksum_address(address)


def generate_mnemonic(num_words: int = 24) -> str:
    mnemonic = Mnemonic("english")  # type: ignore[no-untyped-call]
    return mnemonic.generate(num_words=num_words)


def generate_accounts(  # nosec
    mnemonic: str,
    passphrase: str = "",
    path_prefix: str = "m/44'/60'/0'/0",
    limit: int = 12,
) -> list[NewAccount]:
    result: list[NewAccount] = []
    for i in range(limit):
        path = f"{path_prefix}/{i}"
        acc = Account.from_mnemonic(mnemonic=mnemonic, account_path=path, passphrase=passphrase)
        result.append(NewAccount(path, acc.address, acc.key.hex()))
    return result


def private_to_address(private_key: str) -> str | None:
    """returns address in lower case"""
    # noinspection PyBroadException
    try:
        return key_api.PrivateKey(decode_hex(private_key)).public_key.to_address().lower()
    except Exception:
        return None


def create_private_keys_dict(private_keys: list[str]) -> dict[str, str]:  # address in lower
    result = {}
    for private_key in private_keys:
        address = private_to_address(private_key)
        if address is None:
            raise ValueError("wrong private key")
        result[address.lower()] = private_key
    return result


def is_private_key(private_key: str) -> bool:
    # noinspection PyBroadException
    try:
        key_api.PrivateKey(decode_hex(private_key)).public_key.to_address()
        return True
    except Exception:
        return False



from mm_eth import tx


def test_sign_legacy_tx(private_0, address_1):
    raw_tx = "0xf8601f0b169458487485c3858109f5a37e42546fe87473f79a4b218312399926a00739cb05ce6b1370b4dd53745e51202c2416427d47f60dd19cedbbec1ae8d5d0a064eea36637d9bd09b6e2efa8d20a906fd0146a6798a1b4db29457b6e9539862b"  # noqa
    res = tx.sign_legacy_tx(
        nonce=31,
        gas_price=11,
        gas=22,
        private_key=private_0,
        chain_id=1,
        data="0x123999",
        value=33,
        to=address_1,
    )
    assert res.tx_hash == "0x0400cadbd22ce24ef98ffa0ffe337997758edbde5e4faebe6cda63d8246f545c"
    assert res.raw_tx == raw_tx


def test_decode_raw_tx(address_1):
    raw_tx = "0xf8601f0b169458487485c3858109f5a37e42546fe87473f79a4b218312399926a00739cb05ce6b1370b4dd53745e51202c2416427d47f60dd19cedbbec1ae8d5d0a064eea36637d9bd09b6e2efa8d20a906fd0146a6798a1b4db29457b6e9539862b"  # noqa
    res = tx.decode_raw_tx(raw_tx)
    assert res.tx_hash == "0x0400cadbd22ce24ef98ffa0ffe337997758edbde5e4faebe6cda63d8246f545c"
    assert res.to == address_1


def test_encode_raw_tx_with_signature():
    raw_tx = "0xf8601f0b16949e8daea6fed7e024d9d4c802115a81e908905a8e218312399925a0bd1601a62708068a486c1b0fc793f368fbe629823cb9fd861c9c3c2a37aa7ca6a07dfc3abd01cbfa177fb77595b883ee1062d5146bc006b2c8afa6199defac1dc1"  # noqa
    decoded_tx = tx.decode_raw_tx(raw_tx)
    res = tx.encode_raw_tx_with_signature(
        nonce=decoded_tx.nonce,
        gas_price=decoded_tx.gas_price,
        gas=decoded_tx.gas,
        v=decoded_tx.v,
        r=decoded_tx.r,
        s=decoded_tx.s,
        data=decoded_tx.data,
        value=decoded_tx.value,
        to=decoded_tx.to,
    )
    assert res == raw_tx
